# ESA (C) 2000-2021
# 
# This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
#
#    SAS is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    SAS is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with SAS.  If not, see <http://www.gnu.org/licenses/>.

# pyeslewsearch.py

import os
import sys
import time
import pysas.pyutils.pyutils as pyutils
from logger import get_logger
from astropy.io import fits
from astropy.io import ascii
from astropy.table import Table, unique, QTable
from astropy import units as u
import numpy as np
import re
import glob


logger = get_logger('pyeslewsearch')

from .version import VERSION
from pysas import SAS_RELEASE, SAS_AKA

__version__ = f'pyeslewsearch (pyeslewchain-{VERSION}) [{SAS_RELEASE}-{SAS_AKA}]'


def pyeslewsearch():
    """
    Main program of the task. The function will be called when from the
    command line.
    """

    t_start = time.time()

    coldesc = ['DSSNAME', 'SRCNAME', 'XIMNAME']

    expmaps = []

    for i in os.listdir(os.getcwd()):
        if 'EXPMAP8' in i:
            expmaps.append(i)

    if len(expmaps) == 0:
        logger.error('Empty exposure map list.')
        sys.exit(0)

    expmaps = sorted(expmaps)

    for expmap in expmaps:
        for band_code in ('6', '7', '8'):
            expin = expmap[0:23] + band_code + expmap[24:30]
            imin = expmap[0:17] + 'IMAGE_' + band_code + expmap[24:30]
            srclst = expmap[0:17] + 'SRCLST' + band_code + expmap[24:30]

            ecf = get_ECF(band_code)
            n_events = num_events(imin)

            if n_events > 0:
                eslewsearch_guts(imin, expin, srclst, ecf)
            else:
                logger.info('Empty image ({}).', imin)
    
    b6list = expmaps[0][0:17] + 'OMSRLI6000.FIT'
    b7list = expmaps[0][0:17] + 'OMSRLI7000.FIT'
    b8list = expmaps[0][0:17] + 'OMSRLI8000.FIT'
   
    srclst6_files = glob.glob('*SRCLST6*')

    if len(srclst6_files) != 0:
        logger.info('Working on source list files for band 6.')

        status = pyutils.merge_fits(srclst6_files, [1], b6list, columns = None, new_ext = None)
        if status:
            logger.error("Failed to merge band 6 lists")

    srclst7_files = glob.glob('*SRCLST7*')
    
    if len(srclst7_files) != 0:
        status = pyutils.merge_fits(srclst7_files, [1], b7list, columns = None, new_ext = None)
        if status:
            logger.error("Failed to merge band 7 lists")

    srclst8_files = glob.glob('*SRCLST8*')

    if len(srclst8_files) != 0:
        status = pyutils.merge_fits(srclst8_files, [1], b8list, columns = None, new_ext = None)
        if status:
            logger.error("Failed to merge band 8 lists")

    logger.info('Sizes of each band: B6: {}, B7: {} B8: {}.', len(srclst6_files), len(srclst7_files), len(srclst8_files))
    # cleaning up...
    logger.info('Cleaning up...')
    srclistrm = glob.glob('*SRCLST*')

    for file_rm in srclistrm:
        os.remove(file_rm)

    # Append the different band lists into a single list (no source merging)
    create_single_list(b6list, b7list, b8list)
    
    del_list = glob.glob('box*.ds')
    if os.path.isfile('detmask.fits'):
        del_list.append('detmask.fits')
    if os.path.isfile('splinemap.ds'):
        del_list.append('splinemap.ds')

    for f in del_list:
        os.remove(f)

    t_stop = time.time()

    logger.info('eslewsearch completed in {} seconds.', round(t_stop - t_start, 2))


def add_name_cols(mlist, image):
    """
    Adds a column into the source list.

    Args:
        mlist: the original list.
        image: the image file.
    """

    logger.debug('Running add_name_cols with source list {} and image {}.', mlist, image)
    
    coord_string = make_coord_strings(mlist)

    fstrings = make_file_string(image, coord_string)
    # Write strings to data file

    temp_col_fits = 'temp_col_data.fits'

    # Convert name strings into FITS format

    logger.debug('Running create_fits_from_data.')
    coldesc = ['DSSNAME', 'SRCNAME', 'XIMNAME']
    
    coldata = fstrings
    coldata = np.array(coldata).T.tolist()
    status = pyutils.create_fits_from_data(coldata, coldesc, temp_col_fits)
    
    if status:
        logger.error('Error while running create_fits_from_data')
    
    status = pyutils.add_column_to_fits(mlist, 1, coldesc, coldata)
    
    if status:
        logger.error('Error while running add_column.')
        sys.exit(0)
   

def make_file_string(image, coords):
    """
    Creates a string containing basic information from the file.

    Args: 
        image: the name of the image file.
        coords: the coordinates corresponding to the file.

    Output:
        fstring: the formatted string.
    """

    logger.debug('Running make_file_string with image {}.', image)

    obsid = image[1:11]
    rev = image[2:6]
    fstrings = []
    
 # Loop over each coordinate string and create a string for import
    for coord in coords:
        dsstring = "ds" + obsid + "_" + coord[0] + coord[1] + ".ds"
        srcname = "xs" + rev + "_" + obsid + "_" + coord[0] + coord[1] + ""
        fstring = (dsstring, srcname, "" + image + "")
        fstrings.append(fstring)
    
    return fstrings


def convert_name(srcin):
    """
    Adds the individual band list into one list.

    Args:
        srcin: the information from the source string.

    Output:
        srcout: the name for the source.
    """

    srcin = srcin.replace("'", '')
    srcout = "XMMSL_J" + srcin[18:].replace(':', '')
    return srcout


def create_single_list(b6list, b7list, b8list):
    """
    Creates a single list containing information from each energy bband.

    Args:
        b6list: the list of the sources in the 6 band.
        b7list: the list of the sources in the 7 band.
        b8list: the list of the sources in the 8 band.
    """

    logger.debug('Running create_single_list...')

    b6temp = 'temp_b6.dat'
    b7temp = 'temp_b7.dat'
    b8temp = 'temp_b8.dat'
    tempmerge = 'temp_merged.dat'
    mergedlist = b8list[0:19] + 'SSLI0000.FIT'

    # Combine into a single text file with separate columns for each band

    if os.path.isfile(b6list):
        # Dump files into ascii
        with fits.open(b6list) as ff6:
            t = Table(ff6[1].data) 
            t = t['SRCNAME', 'RA', 'DEC', 'RADEC_ERR', 'SCTS', 'SCTS_ERR', 'EXT', 'EXT_ERR', 'DET_ML', 'BG_MAP', 'RATE', 'RATE_ERR','FLUX', 'FLUX_ERR',
            'XIMNAME', 'ID_INST']
            t.write(b6temp, format = 'ascii')

        logger.info('Working with B6...')

        with open(b6temp) as b6_tf:
            rows_b6 = []

            for line in b6_tf.readlines():
                aux_list = line.split(" ")
                # Add useful records (non-titles and with inst!=0) to output text file
                ii = len(aux_list)
                
                if ii > 0:
                    #for k in range(0, ii):
                        #try:
                            #aux_list[k] = str(np.float32(aux_list[k]))
                        #except:
                            #continue
                    if 'xs' in aux_list[0] and aux_list[15] != 0:
                        # Convert source name to XMMSL format
                        cname = convert_name(aux_list[0])
                        # Write output record\
                        row = [cname, aux_list[1], aux_list[2], aux_list[3], aux_list[4], aux_list[5], np.nan, np.nan, np.nan, np.nan,\
                            aux_list[6], aux_list[7], np.nan, np.nan, np.nan, np.nan, aux_list[8], \
                            np.nan, np.nan, aux_list[9], np.nan, np.nan, aux_list[10], aux_list[11], \
                            np.nan, np.nan, np.nan, np.nan, aux_list[12], aux_list[13], \
                            np.nan, np.nan, np.nan, np.nan, aux_list[14]]

                        rows_b6.append(row)

    else:
        logger.error('Could not open the B6 list.')

    logger.info('Working with B7...')
    
    if os.path.isfile(b7list):
        with fits.open(b7list) as ff:
            t = Table(ff[1].data)
            t = t['SRCNAME', 'RA', 'DEC', 'RADEC_ERR', 'SCTS', 'SCTS_ERR', 'EXT' , 'EXT_ERR' ,'DET_ML', 'BG_MAP', 'RATE', 'RATE_ERR','FLUX', 'FLUX_ERR', 'XIMNAME', 'ID_INST']
            ascii.write(t, b7temp, overwrite = True)

        with open(b7temp) as b7_tf:
            rows_b7 = []
            for line in b7_tf.readlines():
                aux_list = line.split(' ')
                # Add useful records to output text file
                ii = len(aux_list)
                if ii > 0:
                    #for k in range(0, ii):
                        #try:
                            #aux_list[k] = str(np.float32(aux_list[k]))
                        #except:
                            #continue
                    if 'xs' in aux_list[0] and aux_list[15] != 0:
                        # Convert source name to XMMSL format
                        cname = convert_name(aux_list[0])
                        # Write output record
                        row = [cname, aux_list[1], aux_list[2], aux_list[3], np.nan, np.nan, aux_list[4], aux_list[5], np.nan, np.nan, \
                            np.nan, np.nan, aux_list[6], aux_list[7], np.nan, np.nan, np.nan, aux_list[8], \
                            np.nan, np.nan, aux_list[9], np.nan, np.nan, np.nan, aux_list[10], aux_list[11], \
                            np.nan, np.nan, np.nan, np.nan, aux_list[12], aux_list[13], \
                            np.nan, np.nan, aux_list[14]]

                        rows_b7.append(row)


    else:
        logger.error('Could not open the B7 list.')

    logger.info('Working with B8...')

    if os.path.isfile(b8list):
        with fits.open(b8list) as ff:
            t = Table(ff[1].data)
            t = t['SRCNAME', 'RA', 'DEC', 'RADEC_ERR', 'SCTS', 'SCTS_ERR', 'EXT' , 'EXT_ERR' ,'DET_ML', 'BG_MAP', 'RATE', 'RATE_ERR','FLUX', 'FLUX_ERR', 'XIMNAME', 'ID_INST']
            ascii.write(t, b8temp, overwrite = True)

        with open(b8temp) as b8_tf:
            rows_b8 = []
            for line in b8_tf.readlines():
                aux_list = line.split(" ")
                # Add useful records to output text file
                ii = len(aux_list)
                if ii > 0:
                    #for k in range(0, ii):
                        #try:
                            #aux_list[k] = str(np.float32(aux_list[k]))
                        #except:
                            #continue
                    if 'xs' in aux_list[0] and aux_list[15] != 0:
                        # Convert source name to XMMSL format
                        cname = convert_name(aux_list[0])
                        # Write output record
                        row = [cname, aux_list[1], aux_list[2], aux_list[3], np.nan, np.nan, np.nan, np.nan, aux_list[4], aux_list[5], np.nan, np.nan, \
                            np.nan, np.nan, aux_list[6], aux_list[7], np.nan, np.nan,  aux_list[8], \
                            np.nan, np.nan, aux_list[9], np.nan, np.nan, np.nan, np.nan, aux_list[10], aux_list[11], \
                            np.nan, np.nan, np.nan, np.nan, aux_list[12], aux_list[13], aux_list[14]]

                        rows_b8.append(row)

    else:
        logger.error('Could not open the B8 list.')

    logger.info('Write the merged text file into a FITS file.')
    
    labels = make_final_cols()

    logger.info('Joining band files...')

    start_b6_data = np.array(rows_b6).T.tolist()
    status = pyutils.create_fits_from_data(start_b6_data, labels, 'b6band.temp', 'SRCLIST')
    
    logger.debug('Creating temp 6')
    start_b7_data = np.array(rows_b7).T.tolist()
    status = pyutils.create_fits_from_data(start_b7_data, labels, 'b7band.temp', 'SRCLIST')
    start_b8_data = np.array(rows_b8).T.tolist()
    status = pyutils.create_fits_from_data(start_b8_data, labels, 'b8band.temp', 'SRCLIST')

    temp_band_files = ('b6band.temp', 'b7band.temp', 'b8band.temp')
    status = pyutils.merge_fits(temp_band_files, [1], mergedlist, columns = None, new_ext = None)

    for f in temp_band_files:
        if os.path.isfile(f):
            os.remove(f)

    #status = pyutils.add_row_to_fits(mergedlist, 'SRCLIST', rows_b7)
    #if status:
    #    logger.log('error', 'Error while running add_rows_to_fits.')
    
    #status = pyutils.add_row_to_fits(mergedlist, 'SRCLIST', rows_b8)
    
    #if status:
    #    logger.log('error', 'Error while running add_row_to_fits.')

    # Finally sort the list on RA, deleting additional entries.

    with fits.open(mergedlist) as ff:
        t = QTable(ff[1].data)
        t = unique(t, 'RA')
        t['RA'] = t['RA'].astype('float')
        t['DEC'] = t['DEC'].astype('float')
        t.sort('RA')

    qt = add_custom_format(t)
    qt = add_units_table(qt)

    try:
        qt.write(mergedlist, format = 'fits', overwrite = True)
    except:
        logger.error('Could not save the file.')
        sys.exit(0)

    # Change the extension
    with fits.open(mergedlist, 'update') as ef:        
        ef[1].name = 'SRCLIST'
        ef.flush()
   

    add_comments(mergedlist)
    add_units_band(b6list)
    add_units_band(b7list)
    add_units_band(b8list)

    # Clean up but do not throw error if no files to clean
    if os.path.isfile(b6temp):
        os.remove(b6temp)
    if os.path.isfile(b7temp):
        os.remove(b7temp)
    if os.path.isfile(b8temp):
        os.remove(b8temp)
    if os.path.isfile('temp_col_data.fits'):
        os.remove('temp_col_data.fits')


def add_custom_format(t):
    """
    Changes the type of the values for the input data.

    Args:
        t: the input table.
    Output:
        t (corrected).
    """

    aux_srcname = np.chararray(len(t['SRCNAME']), itemsize = 40, unicode = False)
    aux_srcname[:] = t['SRCNAME']
    t['SRCNAME'] = aux_srcname
    t['RA'] = np.array(t['RA'], dtype = 'float64')
    t['DEC'] =  np.array(t['DEC'], dtype = 'float64')
    t['RADEC_ERR'] = np.array(t['RADEC_ERR'], dtype =  'float64')
    t['SCTS_B6'] = np.array(t['SCTS_B6'], dtype = 'float32')
    t['SCTS_B6_ERR'] = np.array(t['SCTS_B6_ERR'], dtype = 'float32')
    t['SCTS_B7'] = np.array(t['SCTS_B7'],dtype = 'float32')
    t['SCTS_B7_ERR'] = np.array(t['SCTS_B7_ERR'], dtype = 'float32')
    t['SCTS_B8'] = np.array(t['SCTS_B8'], dtype = 'float32')
    t['SCTS_B8_ERR'] = np.array(t['SCTS_B8_ERR'], dtype = 'float32')
    t['EXT_B6'] = np.array(t['EXT_B6'], dtype = 'float32')
    t['EXT_B6_ERR'] = np.array(t['EXT_B6_ERR'], dtype = 'float32')
    t['EXT_B7'] = np.array(t['EXT_B7'], dtype = 'float32')
    t['EXT_B7_ERR'] = np.array(t['EXT_B7_ERR'], dtype = 'float32')
    t['EXT_B8'] = np.array(t['EXT_B8'], dtype = 'float32')
    t['EXT_B8_ERR'] = np.array(t['EXT_B8_ERR'], dtype = 'float32')
    t['DET_ML_B6'] = np.array(t['DET_ML_B6'], dtype = 'float32')
    t['DET_ML_B7'] = np.array(t['DET_ML_B7'], dtype = 'float32')
    t['DET_ML_B8'] = np.array(t['DET_ML_B8'], dtype = 'float32')
    t['BG_MAP_B6'] = np.array(t['BG_MAP_B6'], dtype = 'float32')
    t['BG_MAP_B7'] = np.array(t['BG_MAP_B7'], dtype = 'float32')
    t['BG_MAP_B8'] = np.array(t['BG_MAP_B8'], dtype = 'float32')
    t['RATE_B6'] = np.array(t['RATE_B6'], dtype = 'float32')
    t['RATE_B6_ERR'] = np.array(t['RATE_B6_ERR'], dtype = 'float32')
    t['RATE_B7'] = np.array(t['RATE_B7'], dtype = 'float32')
    t['RATE_B7_ERR'] = np.array(t['RATE_B7_ERR'], dtype = 'float32')    
    t['RATE_B8'] = np.array(t['RATE_B8'], dtype = 'float32')
    t['RATE_B8_ERR'] = np.array(t['RATE_B8_ERR'], dtype = 'float32')
    t['FLUX_B6'] = np.array(t['FLUX_B6'], dtype = 'float32')
    t['FLUX_B6_ERR'] = np.array(t['FLUX_B6_ERR'], dtype = 'float32') 
    t['FLUX_B7'] = np.array(t['FLUX_B7'], dtype = 'float32')
    t['FLUX_B7_ERR'] = np.array(t['FLUX_B7_ERR'], dtype = 'float32')
    t['FLUX_B8'] = np.array(t['FLUX_B8'], dtype = 'float32')
    t['FLUX_B8_ERR'] = np.array(t['FLUX_B8_ERR'], dtype = 'float32')
    aux_ximname = np.chararray(len(t['XIMNAME']), itemsize = 40, unicode = False)
    aux_ximname[:] = t['XIMNAME']
    t['XIMNAME'] = aux_ximname

    return t


def add_units_band(blist):
    """
    Moves the input table into a Quantity table object and gets the proper units for each column. Applied to band fits files.

    Args:
        blist: the FITS file of the band.

    Output:
        1 when finished.
    """

    logger.debug('Running add_units_band with {}.', blist)

    with fits.open(blist) as bl:
        qt = QTable(bl[1].data)

    qt = add_custom_format_band(qt)

    try:
        qt['SCTS'].unit = u.ct
        qt['SCTS_ERR'].unit = u.ct
        qt['X_IMA'].unit = u.pix
        qt['X_IMA_ERR'].unit = u.pix
        qt['Y_IMA'].unit = u.pix
        qt['Y_IMA_ERR'].unit = u.pix
        qt['EXT'].unit = u.pix
        qt['EXT_ERR'].unit = u.pix
        qt['BG_MAP'].unit = u.ct / u.pix
        qt['EXP_MAP'].unit = u.s
        qt['FLUX'].unit = u.erg / u.s / (u.cm * u.cm) 
        qt['FLUX_ERR'].unit = u.erg / u.s / (u.cm * u.cm)
        qt['RATE'].unit = u.ct / u.s
        qt['RATE_ERR'].unit = u.ct / u.s
        qt['RA'].unit = u.deg
        qt['DEC'].unit = u.deg
        qt['RADEC_ERR'].unit = u.arcsec
        qt['LII'].unit = u.deg
        qt['BII'].unit = u.deg
        qt['RAWX'].unit = u.pix
        qt['RAWY'].unit = u.pix
        qt['OFFAX'].unit = u.arcmin
        qt['CUTRAD'].unit = u.pix
        qt['DIST_NN'].unit = u.arcsec
    except:
        logger.warning('Could not fit all the data in {}.', blist)
    

    qt.write(blist, format = 'fits', overwrite = True)

    return 1


def add_custom_format_band(t):
    """
    Corrects the format of the input data table.

    Args:
        t: the data table for bands B6, B7 and B8.

    Output:
        t (corrected)
    """

    logger.debug('Running add_custom_format...')
    
    t['ML_ID_SRC'] = np.array(t['ML_ID_SRC'], dtype = 'int32')
    t['BOX_ID_SRC'] = np.array(t['BOX_ID_SRC'], dtype = 'int32')
    t['ID_INST'] = np.array(t['ID_INST'], dtype = 'int32')
    t['ID_BAND'] = np.array(t['ID_BAND'], dtype = 'int32')
    t['ID_CLUSTER'] = np.array(t['ID_CLUSTER'], dtype = 'int32')
    t['SCTS'] = np.array(t['SCTS'], dtype = 'float32')
    t['SCTS_ERR'] = np.array(t['SCTS_ERR'], dtype = 'float32')
    t['X_IMA'] = np.array(t['X_IMA'], dtype = 'float32')
    t['X_IMA_ERR'] = np.array(t['X_IMA_ERR'], dtype = 'float32')
    t['Y_IMA'] = np.array(t['Y_IMA'], dtype = 'float32')
    t['Y_IMA_ERR'] = np.array(t['Y_IMA_ERR'], dtype = 'float32')
    t['EXT'] = np.array(t['EXT'], dtype = 'float32')
    t['EXT_ERR'] = np.array(t['EXT_ERR'], dtype = 'float32')
    t['DET_ML'] = np.array(t['DET_ML'], dtype = 'float32')
    t['EXT_ML'] = np.array(t['EXT_ML'], dtype = 'float32')
    t['BG_MAP'] = np.array(t['BG_MAP'], dtype = 'float32')
    t['EXP_MAP'] = np.array(t['EXP_MAP'], dtype = 'float32')
    t['FLUX'] = np.array(t['FLUX'], dtype = 'float32')
    t['FLUX_ERR'] = np.array(t['FLUX_ERR'], dtype = 'float32')
    t['RATE'] = np.array(t['RATE'], dtype = 'float32')
    t['RATE_ERR'] = np.array(t['RATE_ERR'], dtype = 'float32')
    t['RA'] = np.array(t['RA'], dtype = 'float64')
    t['DEC'] = np.array(t['DEC'], dtype = 'float64')
    t['RADEC_ERR'] = np.array(t['RADEC_ERR'], dtype = 'float32')
    t['LII'] = np.array(t['LII'], dtype = 'float64')
    t['BII'] = np.array(t['BII'], dtype = 'float64')
    t['RAWX'] = np.array(t['RAWX'], dtype = 'int32')
    t['RAWY'] = np.array(t['RAWY'], dtype = 'int32')
    t['OFFAX'] = np.array(t['OFFAX'], dtype = 'float32')
    t['CCDNR'] = np.array(t['CCDNR'] , dtype = 'int32')
    t['HR1'] = np.array(t['HR1'], dtype = 'float32')
    t['HR1_ERR'] = np.array(t['HR1_ERR'], dtype = 'float32')
    t['HR2'] = np.array(t['HR2'], dtype = 'float32')
    t['HR2_ERR'] = np.array(t['HR2_ERR'], dtype = 'float32')
    t['HR3'] = np.array(t['HR3'], dtype = 'float32')
    t['HR3_ERR'] = np.array(t['HR3_ERR'], dtype = 'float32')
    t['HR4'] = np.array(t['HR4'], dtype =  'float32')
    t['HR4_ERR'] = np.array(t['HR4_ERR'], dtype = 'float32')
    t['CUTRAD'] = np.array(t['CUTRAD'], dtype = 'float32')
    t['MASKFRAC'] = np.array(t['MASKFRAC'], dtype = 'float32')
    t['EEF'] = np.array(t['EEF'], dtype = 'float32')
    t['VIGNETTING'] = np.array(t['VIGNETTING'], dtype = 'float32')
    t['ONTIME'] = np.array(t['ONTIME'], dtype = 'float32')
    t['PILEUP'] = np.array(t['PILEUP'], dtype = 'float32')
    t['DIST_NN'] = np.array(t['DIST_NN'], dtype = 'float32')
    aux_flag = np.chararray(len(t['FLAG']), itemsize = 12, unicode = False)
    aux_flag[:] = t['FLAG']
    t['FLAG'] = aux_flag
    aux_DSSNAME = np.chararray(len(t['DSSNAME']), itemsize = 40, unicode = False)
    aux_DSSNAME[:] = t['DSSNAME']
    t['DSSNAME'] = aux_DSSNAME
    aux_srcname = np.chararray(len(t['SRCNAME']), itemsize = 40, unicode = False)
    aux_srcname[:] = t['SRCNAME']
    t['SRCNAME'] = aux_srcname
    aux_ximname = np.chararray(len(t['XIMNAME']), itemsize = 40, unicode = False)
    aux_ximname[:] = t['XIMNAME']
    t['XIMNAME'] = aux_ximname
    #DSSNAME   str40
    #SRCNAME   str40
    #XIMNAME   str40

    return t


def add_units_table(qt):
    """
    Moves the input table into a Quantity table object and gets the proper units for each column.

    Args:
        qt: the input table without units.

    Output:
        qt: the Quantity Table with proper units.
    """
    
    logger.log('debug', 'Adding units to main table...')
    
    try:
        qt['RA'] = qt['RA'] * u.deg
        qt['DEC'].unit = u.deg
        qt['RADEC_ERR'].unit = u.arcsec
        qt['EXT_B6'].unit = u.pix
        qt['EXT_B6_ERR'].unit = u.pix
        qt['EXT_B7'].unit = u.pix
        qt['EXT_B7_ERR'].unit = u.pix
        qt['EXT_B8'].unit = u.pix
        qt['EXT_B8_ERR'].unit = u.pix
        qt['BG_MAP_B6'].unit = u.ct / u.pix
        qt['BG_MAP_B7'].unit = u.ct / u.pix
        qt['BG_MAP_B8'].unit = u.ct / u.pix
        qt['RATE_B6'].unit = u.ct / u.s
        qt['RATE_B6_ERR'].unit = u.ct / u.s
        qt['RATE_B7'].unit = u.ct / u.s
        qt['RATE_B7_ERR'].unit = u.ct / u.s
        qt['RATE_B8'].unit = u.ct / u.s
        qt['RATE_B8_ERR'].unit = u.ct / u.s
        qt['FLUX_B6'].unit = u.erg / u.s / (u.cm * u.cm)
        qt['FLUX_B6_ERR'].unit = u.erg / u.s / (u.cm * u.cm)
        qt['FLUX_B7'].unit = u.erg / u.s / (u.cm * u.cm)
        qt['FLUX_B7_ERR'].unit = u.erg / u.s / (u.cm * u.cm)
        qt['FLUX_B8'].unit = u.erg / u.s / (u.cm * u.cm)
        qt['FLUX_B8_ERR'].unit = u.erg / u.s / (u.cm * u.cm)
    except:
        logger.warning('An error ocurred while adding the units to the final source list. It may be incomplete.')

    return qt


def num_events(image):
    """
    Gets the total number of events in a given image.

    Args:
        image: the path or FITS image to be evaluated.
    Output:
        sum: the total sum returned by imgstat.
    """

    stat_image = pyutils.imgstat(image, 0, None)

    return stat_image[4]


def make_coord_strings(mlist):
    """
    Reads coordinates and converts them into HMS,dms.

    Args:
        mlist: list to check coordinates.

    Output:
        new_coords: a list of the tuples of the new coordinates.
    """

    new_coords = []
    coords = pyutils.pydump(mlist, 1, ['RA', 'DEC'])

    for row in coords:
        ra_o = float(row[0])
        dec_o = float(row[1])
        values_coords = (convert_RA(ra_o), convert_DEC(dec_o))
        new_coords.append(values_coords)
    return new_coords


def make_final_cols():
    """
    Returns the main labels to use in the column to be formatted.

    Output:
        labels: the labels to be used.
    """
    
    labels = ['SRCNAME',
 'RA',
 'DEC',
 'RADEC_ERR',
 'SCTS_B6',
 'SCTS_B6_ERR',
 'SCTS_B7',
 'SCTS_B7_ERR',
 'SCTS_B8',
 'SCTS_B8_ERR',
 'EXT_B6',
 'EXT_B6_ERR',
 'EXT_B7',
 'EXT_B7_ERR',
 'EXT_B8',
 'EXT_B8_ERR',
 'DET_ML_B6',
 'DET_ML_B7',
 'DET_ML_B8',
 'BG_MAP_B6',
 'BG_MAP_B7',
 'BG_MAP_B8',
 'RATE_B6',
 'RATE_B6_ERR',
 'RATE_B7',
 'RATE_B7_ERR',
 'RATE_B8',
 'RATE_B8_ERR',
 'FLUX_B6',
 'FLUX_B6_ERR',
 'FLUX_B7',
 'FLUX_B7_ERR',
 'FLUX_B8',
 'FLUX_B8_ERR',
 'XIMNAME']

    return labels


def eslewsearch_guts(image, expmap, mlist, ecf):
    """
    Runs several SAS' tasks to create additional files. Will quit if the aforementioned taasks run into any error (check the log file).

    Args:
        image: the image to manipulate.
        expmap: the exposure map for the image.
        mlist:
        ecf: the energy correction factor.
    """

    logger.debug('Running eslewsearch_guts with {}, {}, {}.', image, expmap, mlist)

    # Set constants and thresholds
    pimin = 1499 # USe 1.5 keV for the PSF in all energy-band images!!
    pimax = 1501
    eboxlthr = 8
    eboxmthr = 10
    emlthr = 8
    mask = 'detmask.fits'
    boxlist_l = 'boxlist_l.ds'
    boxlist_m = 'boxlist_m.ds'
    bckgnd = 'splinemap.ds'

    # Run emask
    status_code = os.system("emask expimageset={0} detmaskset={1} threshold1=1e-5 threshold2=0.5".format(expmap, mask))
    
    if status_code:
        logger.error('Could not run emask.')
        sys.exit(0)

    # Run eboxdetect (local mode)
    status_code = os.system("eboxdetect boxlistset={0} boxsize=5 expimagesets={1} imagesets={2} likemin={3} nruns=3 pimax={4} pimin={5} usemap=no usematchedfilter=no withdetmask=yes detmasksets={6} withexpimage=yes withoffsets=no obsmode='slew'".format(boxlist_l, expmap, image, eboxlthr, pimax, pimin, mask))
    
    if status_code:
        logger.error("Failure in eboxdetect (local) on {}", image)
        sys.exit(0)

    # Create background map
    status_code = os.system("esplinemap bkgimageset={0} boxlistset={1} withcheese=no withdetmask=yes detmaskset={2} excesssigma=3 idband=1 imageset={3} mlmin=1 nfitrun=3 nsplinenodes=10 scut=0.005 withexpimage=no".format(bckgnd, boxlist_l, mask, image))
    if status_code:
        logger.error("Failure in esplinemap on {}.", image)
        sys.exit(0)

    # Run eboxdetect (map mode)
    status_code = os.system("eboxdetect bkgimagesets={0} boxlistset={1} boxsize=5 withexpimage=yes expimagesets={2} imagesets={3} likemin={4} nruns=3 pimax={5} pimin={6} usemap=yes usematchedfilter=no withdetmask=yes detmasksets={7} withoffsets=no obsmode='slew'".format(bckgnd, boxlist_m, expmap, image, eboxmthr, pimax, pimin, mask))
    
    if status_code:
        logger.error("Failure in eboxdetect (map) on {}", image)
        sys.exit(0)

    # Run emldetect
    status_code = os.system("emldetect bkgimagesets={0} boxlistset={1} determineerrors=yes dmlextmin=2 ecf={2} ecut=36 withexpimage=yes expimagesets={3} fitcounts=yes fitextent=yes fitposition=yes imagesets={4} mllistset={5} mlmin={6} pimax={7} pimin={8} scut=0.9 usecalpsf=yes useevents=no withhotpixelfilter=no withoffsets=no withxidband=no nmaxfit=1 nmulsou=1 psfmodel='slew'".format(bckgnd, boxlist_m, ecf, expmap, image, mlist, emlthr, pimax, pimin))
    
    if status_code:
        logger.error("Failure in emldetect on {}", image)
        sys.exit(0)

    if os.path.isfile(mlist):
        add_name_cols(mlist, image)


def convert_RA(ra):
    """
    Change the format of the RA parameter and returns it as a formatted string.

    Args:
        ra: the numerical value of RA.

    Output:
        hms: the formatted RA string.
    """

    hrs = int(ra / 15)
    mins = int((ra - hrs * 15.0) * 60.0 / 15.0)
    secs = (ra - (hrs * 15.0 + mins * 15.0 / 60.0)) * 3600.0 / 15.0

    if secs > 59.95:
        secs = 0
        mins = mins + 1

    if mins == 60:
        mins = 0
        hrs = hrs + 1

    if hrs == 24:
        hrs = 0

    #hms = '{0}:{1}:{2}'.format(str(round(hrs, 2)).zfill(2), str(round(mins, 2)).zfill(2), round(secs, 4))

    hms = '{0}:{1}:{2}'.format(str(round(hrs, 2)).zfill(2), str(round(mins, 2)), round(secs, 1))

    return hms


def convert_DEC(dec):
    """
    Converts the dec into a formatted string.

    Args:
        dec: the dec value as a real number.

    Output:
        degstring:
    """

    deg = int(dec)
    mins = int((dec - deg) * 60.0)
    secs = int(abs(dec - (deg + mins / 60.0)) * 3600.0 + 0.5)
    amins = abs(mins)

    # Sign
    sign = "+"

    if (dec < 0.0):
        sign = "-"

    # Sort out ranges
    if (secs > 59.95):
        secs = 0.0
        amins = amins + 1

    if amins == 60:
        amins = 0.0
        if dec < 0.0:
            deg = deg - 1
        else:
            deg = deg + 1

    degstring = '{0}{1}:{2}:{3}'.format(sign, str(round(abs(deg), 2)).zfill(2), str(round(amins, 2)).zfill(2), str(round(secs, 2)).zfill(2))

    return degstring


def get_ECF(band):
    """
    Returns the energy conversion factor for the corresponding energy band.

    Args:
        band: the energy band (6, 7, 8).

    Output:
        ecf: the energy conversion factor (as a real).
    """

    band = int(band)

    if band == 6:
        ecf = 1.436 # soft energy band (0.2 - 2 keV)
    elif band == 7:
        ecf = 9.144 # hard energy band (2 - 12 keV)
    elif band == 8:
        ecf = 3.159 # total energy band (0.2 - 12 keV)
    else:
        logger.error('The input energy band is not supported ({})', band)
        sys.exit(0)

    ecf = 10 / ecf # for EMLDETECT

    return ecf


def add_comments(mlist):
    """
    Adds the comment info to the first header of the source list.

    Args:
        mlist: the final source list.
    """

    with fits.open(mlist, 'update') as ff:
        for j in ff[1].header:
            field_n = re.findall(r'\d+', j)
            if len(field_n) > 0:
                field_n = field_n[0]
            if ff[1].header[j] == 'RA':
                ff[1].header[j] = ('RA', 'label for field {}'.format(field_n))
            if ff[1].header[j] == 'DEC':
                ff[1].header[j] = ("DEC", 'label for field {}'.format(field_n))
            if ff[1].header[j] == 'RADEC_ERR':
                ff[1].header[j] = ("RADEC_ERR", 'label for field {}'.format(field_n))
            if ff[1].header[j] == "EXT_B6":
                ff[1].header[j] = ("EXT_B6", 'label for field {}'.format(field_n))
            if ff[1].header[j] == 'EXT_B6_ERR':
                ff[1].header[j] = ("EXT_B6_ERR", 'label for field {}'.format(field_n))
            if ff[1].header[j] == 'EXT_B7':
                ff[1].header[j] = ("EXT_B7", 'label for field {}'.format(field_n))
            if ff[1].header[j] == 'EXT_B7_ERR':
                ff[1].header[j] = ("EXT_B7_ERR", 'label for field {}'.format(field_n))
            if ff[1].header[j] == 'EXT_B8':
                ff[1].header[j] = ("EXT_B8", 'label for field {}'.format(field_n))
            if ff[1].header[j] == 'EXT_B8_ERR':
                ff[1].header[j] = ("EXT_B8_ERR", 'label for field {}'.format(field_n))
            if ff[1].header[j] == 'BG_MAP_B6':
                ff[1].header[j] = ("BG_MAP_B6", 'label for field {}'.format(field_n))
            if ff[1].header[j] == 'BG_MAP_B7':
                ff[1].header[j] = ("BG_MAP_B7", 'label for field {}'.format(field_n))
            if ff[1].header[j] == 'BG_MAP_B8':
                ff[1].header[j] = ("BG_MAP_B8", 'label for field {}'.format(field_n))
            if ff[1].header[j] == 'RATE_B6':
                ff[1].header[j] = ("RATE_B6", 'label for field {}'.format(field_n))
            if ff[1].header[j] == 'RATE_B6_ERR':
                ff[1].header[j] = ("RATE_B6_ERR", 'label for field {}'.format(field_n))
            if ff[1].header[j] == 'RATE_B7':
                ff[1].header[j] = ("RATE_B7", 'label for field {}'.format(field_n))
            if ff[1].header[j] == 'RATE_B7_ERR':
                ff[1].header[j] = ("RATE_B7_ERR", 'label for field {}'.format(field_n))
            if ff[1].header[j] == 'RATE_B8':
                ff[1].header[j] = ("RATE_B8", 'label for field {}'.format(field_n))
            if ff[1].header[j] == 'RATE_B8_ERR':
                ff[1].header[j] = ("RATE_B8_ERR", 'label for field {}'.format(field_n))
            if ff[1].header[j] == 'FLUX_B6':
                ff[1].header[j] = ("FLUX_B6", 'label for field {}'.format(field_n))
            if ff[1].header[j] == 'FLUX_B6_ERR':
                ff[1].header[j] = ("FLUX_B6_ERR", 'label for field {}'.format(field_n))
            if ff[1].header[j] == 'FLUX_B7':
                ff[1].header[j] = ("FLUX_B7", 'label for field {}'.format(field_n))
            if ff[1].header[j] == 'FLUX_B7_ERR':
                ff[1].header[j] = ("FLUX_B7_ERR", 'label for field {}'.format(field_n))
            if ff[1].header[j] == 'FLUX_B8':
                ff[1].header[j] = ("FLUX_B8", 'label for field {}'.format(field_n))
            if ff[1].header[j] == 'FLUX_B8_ERR':
                ff[1].header[j] = ("FLUX_B8_ERR", 'label for field {}'.format(field_n))
        ff.flush()


def run(iparsdic):
    print('Executing {}.'.format(__file__))
    pyeslewsearch()




# ESA (C) 2000-2020
# 
# This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
#
#    SAS is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    SAS is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with SAS.  If not, see <http://www.gnu.org/licenses/>.

# pyutils.py

import numpy as np
import subprocess
import os
from astropy.io import fits
import re
import datetime
import sys
from astropy.table import Table


def pyadd(images, outfile):
    """
    Adds several images into a single outfile.
    
    Args:
        images: a tuple or list with the images.
        outfile: the name of the outfile.
    """

    with fits.open(images[0]) as image:
        primary_hdu = image[0]
        sum_im = primary_hdu.data

    for im in range(1, len(images)):
        with fits.open(images[im]) as image1:
            sum_im = sum_im + image1[0].data

    primary_hdu.data = sum_im
    primary_hdu.writeto(outfile, overwrite = True)


def date_time():
    """
    Returns the current date.

    Output:
        the current date time in as string.
    """

    now = datetime.datetime.now()
    return('Current time: {}'.format(now.strftime('%Y - %m - %d -- %H:%M:%s')))


def pyselect(filename, outfile, extension, column, lim_up = ' ', lim_down = ' ',
    return_data = False, exact_match = ''):
    """
    Returns the rows corresponding to a series of conditions imposed by
    given limits for a given fits file. Creates a second fits file with the
    selected rows (second extension, first is the primary extension of the
    original fits file).

    Args:
        filename: the name of the fits file.
        outfile: the name of the output fits file.
        extension: the extension you want to check.
        column: the name of the column to verify.
        lim_up: upper limit.
        lim_down: lower limit.
        return_data: (optional, False by default) returns the selected data as an recarray
        exact_match: whether or not an exact match is required.

    Output:
       selection: creates the file, and depending on return_data, returns the recarray.
    """

    with fits.open(filename) as fits_file:
        hdu_primary = fits_file[0]
        selected_extension = fits_file[extension]

        data_fits = fits_file[extension].data
        if exact_match != '':
            selection = data_fits[data_fits[column] == exact_match]
        else:
            if lim_up != ' ' and lim_down != ' ':
                if lim_down > lim_up:
                    raise ValueError('Upper limit is lower than the lower limit!')
                selection = data_fits[(data_fits[column] > lim_down) & (data_fits[column] < lim_up)]
            elif lim_up != ' ' and lim_down == ' ':
                selection = data_fits[data_fits[column] < lim_up]
            elif lim_up == ' ' and lim_down != ' ':
                selection = data_fits[data_fits[column] > lim_down]
            else:
                raise ValueError('Something went wrong. Check your limits.')

        hdu_primary.writeto(outfile, overwrite = True)

        with fits.open(outfile, 'update') as out_file:
            out_file.append(selected_extension)
            out_file[1].data = selection
            out_file.flush()

    if return_data:
        return selection


def pymodhdu(filename, extension, card, value, comment = ''):
    """
    Modifies the header of a fits file.
    No need for temporary files.
    
    Args:
        filename: name of the file.
        extension: the extension. 0 for primary...
        card: the 'card' that has to be edited.
        value: the value.
        comment: the comment, if any.
    """

    with fits.open(filename, 'update') as fits_file:
        hdr = fits_file[extension].header
        hdr.set(card, value, comment)
        fits_file.flush()


def get_key_word(filename, keyword, extension = '', not_found_output = 'unknown'):
    """
    Search a fits file for a given keyword and returns it.
    
    Args:
        filename: the path of the file to search.
        keyword: the keyword.
        extension (optional): the extension. If none is given it will return the first keyword found that matches.
        not_found_output: the value of the keyword if it couldn't be found in the FITS file.

    Output:
        the value of that given keyword. 'unknown' if not matches were found.
    """

    if isinstance(filename, str):
        if extension == '':
            try:
                with fits.open(filename) as event:
                    for i in range(0, len(event)):
                        if keyword in event[i].header:
                            return event[i].header[keyword]
            except FileNotFoundError:
                return(not_found_output)
        else:
            try:
                with fits.open(filename) as event:
                    hdr = event[extension].header
                    if keyword in hdr:
                        return hdr[keyword]
            except FileNotFoundError:
                return(not_found_output)
    elif isinstance(filename, fits.hdu.hdulist.HDUList):
        if extension == '':
            try:
                for i in range(0, len(filename)):
                    if keyword in filename[i].header:
                        return filename[i].header[keyword]
            except FileNotFoundError:
                return(not_found_output)
        else:
            try:
                hdr = filename[extension].header
                if keyword in hdr:
                    return hdr[keyword]
            except FileNotFoundError:
                return(not_found_output)
    else:
        raise FileNotFoundError('Could not determine the type of the file.')

    return not_found_output


def pydump(fits_file, extension, column):
    """
    Returns the value or values of a wanted column for a fits file.

    Args:
        file: the file name.
        extension: the extension.
        column: the name of the column.

    Output:
        values: values obtained as a float or list of floats. Will be empty if
    there were no matches.
    """

    values = []

    if isinstance(fits_file, str):
        try:
            with fits.open(fits_file) as ev:
                data_table = Table(ev[extension].data)
                data_ev = data_table[column]
                #for i in range(0, len(data_ev)):
                #values.append(data_ev[i].item())
        except KeyError:
            print('Returning empty list. The given file does not contain either the extension or the column.')
            return []
    elif isinstance(fits_file, fits.hdu.hdulist.HDUList):
        try:
            data_table = Table(fits_file[extension].data)
            data_ev = data_table[column]
            #for i in range(0, len(data_ev)):
                #values.append(data_ev[i].item())
        except KeyError:
            print('Returning empty list. The given file does not contain either the extension or the column')
            return []
    else:
        raise TypeError('Unrecognised format for the fits file.')
    return data_ev


def is_notebook():
    """
    Checks whether or not the user is running a Python terminal or a Notebook environment.

    Output:
        True if running on a Jupyter Notebook.
        False if running from terminal, iPython or other systems.
    """

    try:
        shell = get_ipython().__class__.__name__
        if shell == 'ZMQInteractiveShell':
            return True   # Jupyter notebook or qtconsole
        elif shell == 'TerminalInteractiveShell':
            return False  # Terminal running IPython
        else:
            return False  # Other type
    except NameError:
        return False      # Probably standard Python interpreter


def delete_data(fits_file, extension, to_delete, direction):
    """
    Deletes a row or column (or sets of rows or columns) from a given fits file.

    Args:
        fits_file: the fits file. Either the path or an opened fits object.
        extension: the extension to evaluate the data.
        to_delete: the list, string or number of column/row that has to be
    removed.
        direction: 'R' for row, 'C' for column.

    Output:
        None if run into errors.
        If the fits file is a string, will return the string again.
        If the fits file is a HDU object, will return the given HDU object
    modified. You may want to load it again.
    """

    if isinstance(fits_file, str):
        with fits.open(fits_file, 'update') as f:
            try:
                data_in = Table(f[extension].data)
            except KeyError:
                print('Could not open extension.')
                return None
            if direction.upper() == 'R':
                try:
                    if isinstance(to_delete, int):
                        data_in.remove_row(to_delete)
                    elif isinstance(to_delete, list):
                        data_in.remove_rows(to_delete)
                except IndexError:
                    print('Index error while deleting rows. Check the limits.')
                    return None
            elif direction.upper() == 'C':
                if isinstance(to_delete, int) or isinstance(to_delete, str):
                    try:
                        data_in.remove_column(to_delete)
                    except KeyError:
                        print('Could not remove the column {}. It does not exist.'.format(to_delete))
                        return None
                elif isinstance(to_delete, list):
                    try:
                        data_in.remove_columns(to_delete)
                    except KeyError:
                        print('Could not remove the given columns.')
                        return None
            else:
                print('Invalid argument. Please insert C for column(s) or R for row(s)')
                return None
            f[extension].data = None
            f[extension].data = data_in.as_array()
            f.flush()
        return fits_file

    elif isinstance(fits_file, fits.hdu.hdulist.HDUList):
        f = fits_file
        try:
            data_in = Table(f[extension].data)
        except KeyError:
            print('Could not open extension.')
            return None
        if direction.upper() == 'R':
            try:
                if isinstance(to_delete, int):
                    data_in.remove_row(to_delete)
                elif isinstance(to_delete, list):
                    data_in.remove_rows(to_delete)
            except IndexError:
                print('Index error while deleting rows. Check the limits.')
                return None
        elif direction.upper() == 'C':
            if isinstance(to_delete, int) or isinstance(to_delete, str):
                try:
                    data_in.remove_column(to_delete)
                except KeyError:
                    print('Could not remove the column {}. It does not exist.'.format(to_delete))
                    return None
            elif isinstance(to_delete, list):
                try:
                    data_in.remove_columns(to_delete)
                except KeyError:
                    print('Could not remove the given columns. Matches could not be found.')
                    return None
        else:
            print('Invalid argument. Please insert C for column(s) or R for row(s)')
            return None
        f[extension].data = None
        f[extension].data = data_in.as_array()
        return f
    else:
        print('invalid type for the FITS file. Entered {}.'.format(type(fits_file)))
        return None


def create_fits_from_data(data, labels, filename = '', ext_name = None):
    """
    Creates a table object containing the data and the labels passed to its
    arguments. May also create the FITS file if so is specified.

    Args:
        data: list of all the arrays to be included in the table.
        labels: a list of the labels. This list must be of the same lenght as
    the data list.
        filename: empty by default. If the user wants to save the table, this
    argument must contain the name of the file.
        ext_name: the name of the extension. If not provided, the extension name will be empty.

    Output:
        None if the lenght of the lists differ.
        t: the table object.
    """

    if len(data) != len(labels):
        print('Mismatching lenght for the labels and the data columns.')
        return None

    t = Table(data, names = labels)

    if filename != '':
        t.write('{}'.format(str(filename)), format = 'fits', overwrite = True)
    
    if ext_name:
        with fits.open(filename, 'update') as ef:
            ef[1].name = ext_name
            ef.flush()

    return 0


def filter_data(data, nan_values = -999, upper_limit = None, lower_limit = None):
    """
    Allows for easily filter 2D numpy arrays or astropy.Table objects.

    Args:
        data: the data to filter: either a numpy.ndarray or a Table.
        nan_values: the value (or list of values) to be considered nan. NaN
    values will always be considered as such.
        upper_limit: the tolerated upper limit.
        lower_limit: the tolerated lower limit.

    Output:
        data: the filtered data in the same input format. If the input format is not supported, will simply return 1.
    """

    try:
        nan_length = len(nan_values)
    except TypeError:
        nan_length = 0

    if isinstance(data, np.ndarray):
        data = data[~np.isnan(data).any(axis=1)]
        if nan_length == 0:
            data = np.delete(data, np.where(data == nan_values)[0], 0)
        else:
            for i in nan_values:
                data = np.delete(data, np.where(data == i)[0], 0)

        if type(upper_limit) == int or type(upper_limit) == float:
            data = data[data[:,0] < upper_limit, :]

        if type(lower_limit) == int or type(lower_limit) == float:
            data = data[data[:,0] > lower_limit, :]

    elif isinstance(data, Table):
        has_nan = np.zeros(len(data), dtype = bool)
        for col in data.itercols():
            if col.info.dtype.kind == 'f':
                has_nan |= np.isnan(col)
        data = data[~has_nan]
        if nan_length == 0:
            mask = np.logical_or.reduce([c == nan_values for c in data.columns.values()])
            data = data[~mask]
        else:
            for i in nan_values:
                mask = np.logical_or.reduce([c == i for c in data.columns.values()])
                data = data[~mask]

        if type(upper_limit) == int or type(upper_limit) == float:
            mask = np.logical_or.reduce([c > upper_limit for c in data.columns.values()])
            data = data[~mask]
        if type(lower_limit) == int or type(lower_limit) == float:
            mask = np.logical_or.reduce([c < lower_limit for c in data.columns.values()])
            data = data[~mask]

    else:
        print('Format not supported.')
        return 1

    return data


def merge_fits(fits_list, extensions, output_file, columns, new_ext = None):
    """
    Merges the values from a series of given FITS files.

    Args:
        fits_list: the list of fits files to merge.
        extensions: the list of extensions on 1:1 correspondency to the fits file
    list.
        output_file: the name for the output file.
        columns = the columns to take from the data as string (if one) or as a
    list or a tuple (if more than one).
        new_ext: the name of the new extension. If None, the name from the first
    extension of the first file will be used.

    Output:
        0 or 1, depending on whether or not any errors were found.
    """

    data_dict = dict()
    # evaluating columns:
    if isinstance(columns, str):
        data_dict.update({columns : np.array(())})
    elif isinstance(columns, list) or isinstance(columns, tuple):
        for i in columns:
            data_dict.update({i : np.array(())})
    elif columns is None:
        with fits.open(fits_list[0]) as ff:
            names = ff[extensions[0]].data.names
            for i in names:
                data_dict.update({i : np.array(())})
    else:
        print('Unrecognised format.')
        return 1

    size_diff = len(fits_list) - len(extensions)

    for i in range(0, len(fits_list)):
        fits_file = fits_list[i]
        try:
            with fits.open(fits_file) as ffile:
                if i == 0:
                    new_primary = ffile[0]
                    new_base_header = ffile[1]
                    if new_ext == None:
                        new_ext = ffile[1].name
                for key in data_dict.keys():
                    try:
                        if len(extensions) == 1:
                            extension = extensions[0]
                        else:
                            extension = extensions[i]
                        
                        if len(ffile[extension].data) == 0:
                            continue
                        data_key = ffile[extension].data[key]
                        data_in_dict = data_dict[key]
                        data_dict.update({key : np.concatenate((data_in_dict, data_key))})
                    except (IndexError, AttributeError) as e:
                        print('Could not locate {0} in {1}:{2}.'.format(key, fits_file, extension))

        except FileNotFoundError:
            print('Could not open {}.'.format(fits_file))
            return 1

    names = []
    dict_to_values = []

    # Seems like Table does not support dict_keys as a proper list. Converting in a loop:
    for key in data_dict:
        dict_to_values.append(data_dict[key])
        names.append(key)

    new_table = Table(dict_to_values, names = names)
   
    new_table.write(output_file, overwrite = True)

    with fits.open(output_file, 'update') as ff:
        ff[1].name = new_ext
        ff[1].header = new_base_header.header
        #new_hdul = fits.HDUList([new_primary, new_base_header])
        ff.flush()

    return 0


def imgstat(fits_file, lower_limit = None, upper_limit = None, output = None):
    """
    Similar to fimgstat, the function return the statistical information from the image
    in the fits file. Can also be dumped into a txt file.

    Args:
        fits_file:
        lower_limit: the lower value to consider in the statistics.
        upper_limit: the upper value to consider in the statistics.
        output: if specified, will dump the information into the specified file.

    Output:
        mean_image: the mean value of the image.
        std_dev_image: the standard deviation of the image.
        min_val: the minimum value found.
        max_val: the maximum value found.
        sum_val: the total sum of all the values used.
        used_val: the number of values used.
        max_index: the indexes of the maxima.
        min_index: the indexes for the minima.
    """

    if isinstance(fits_file, str):
        try:
            with fits.open(fits_file) as f:
                if f[0].is_image:
                    fits_image = f[0].data
                else:
                    raise AttributeError('The PRIMARY extension of the fits file does not correspond to an image.')
                    sys.exit(0)
        except FileNotFoundError:
            print('Could not open {}.'.format(fits_file))
            sys.exit(0)
    elif isinstance(fits_file, fits.hdu.hdulist.HDUList):
        if fits_file[0].is_image:
            fits_image = fits_file[0].data
        else:
            raise AttributeError('The PRIMARY extension of the fits file does not correspond to an image.')
            sys.exit(0)

    flattened_image = fits_image.flatten()

    # filtering...
    flattened_image = flattened_image[~np.isnan(flattened_image)]
    if lower_limit != None:
        flattened_image = flattened_image[flattened_image >= lower_limit]
    if upper_limit != None:
        flattened_image = flattened_image[flattened_image <= upper_limit]

    mean_image = np.mean(flattened_image)
    std_dev_image = np.std(flattened_image)
    min_val = np.min(flattened_image)
    max_val = np.max(flattened_image)
    used_val = len(flattened_image)
    sum_val = sum(flattened_image)
    max_index = np.asarray(np.where(fits_image == max_val)).T
    min_index = np.asarray(np.where(fits_image == min_val)).T

    if output != None:
        with open(str(output), 'w') as f:
            f.write('Mean value: {}\n'.format(mean_image))
            f.write('Standard deviation: {}\n'.format(std_dev_image))
            f.write('Max value: {}\n'.format(max_val))
            f.write('Min value: {}\n'.format(min_val))
            f.write('Values used in the calculation: {}\n'.format(used_val))
            f.write('Total sum of the values used: {}\n'.format(sum_val))
            f.write('Location of maxima: {}\n'.format(max_index))
            f.write('Location of minima: {}\n'.format(min_index))


    return(mean_image, std_dev_image, min_val, max_val, sum_val, used_val, max_index, min_index)


def add_column_to_fits(fits_file, extension, colname, coldata):
    """
    Adds a column and its data to the given FITS file.

    Args:
        fits_file: the path to the FITS file to be modified.
        extension: in which extension is the data located.
        colname: the name of the column.
        coldata: the data of the column.

    Output:
        0 when finished. 1 otherwise.
    """

    with fits.open(fits_file, 'update') as ff:
        data_t = Table(ff[extension].data)
        primary_head = ff[0].header 
        if isinstance(colname, str):
            try:
                data_t[colname] = coldata
            except:
                try: #insert empty column
                    if len(data_t) == 0 or len(coldata) == 0:
                        data_t[colname] = []
                except:
                    print('Warning: Mismatching sizes in add_column_to_fits.')
                    return 1
        else:
            for i in range(0, len(colname)):
                try:
                    data_t[colname[i]] = coldata[i]
                except:
                    if len(data_t) == 0:
                        try:
                            data_t[colname[i]] = []
                        except:
                            print('Warning: Mismatching size in the {} element (add_column_to_table).'.format(i))
        
        # new file
        data_t.write(fits_file, format = 'fits', overwrite = True)
        with fits.open(fits_file, 'update') as ff2:
            ff2[0].header = primary_head
            ff2.flush()

    return 0


def add_row_to_fits(fits_file, extension, rows):
    """
    Adds a row to the given FITS file in the provided extension.

    Args:
        fits_file: the FITS file to be edited.
        extension: the extension in which the data is found.
        rows: the rows (list format) to add.

    Output:
        0 if completed. 1 otherwise.
    """

    if isinstance(fits_file, str):
        try:
            with fits.open(fits_file, 'update') as ff:
                data = ff[extension].data
                t = Table(data)
                for row in range(0, len(rows)):
                    try:
                        t.add_row(rows[row])
                    except IndexError:
                        print('Could not add row {}.'.format(row))
                        continue
                ff[extension].data = None
                ff[extension].data = t.as_array()
                ff.flush()
        except FileNotFoundError:
            print('Could not open file {0}.'.format(fits_file))
            return 1

    return 0


def sort_fits(fits_file, extension, label, desc = True):
    """
    Rewrites the data of a FITS file after sorting the data of a given extension.

    Args:
        fits_file: the path to the FITS file that has to be modified.
        extension: the extension in which the data is located.
        label: the label of the column to be used. If more than one is provided, use a list of strings.
        desc: bool - if desc(endant) order is wanted.

    Output:
        0 if done correctly; 1 otherwise.
    """
    
    try:
        with fits.open(fits_file, 'update') as ff:
            t = Table(ff[extension].data)
            t.sort(label)
            if not desc:
                t.reverse()
            ff[extension].data = None
            ff[extension].data = t.as_array()
            ff.flush()

    except FileNotFoundError:
        print('Could not open file {}.'.format(fits_file))
        return 1

    return 0



# ESA (C) 2000-2020
# 
# This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
#
#    SAS is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    SAS is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
#
# 
"""parser.py

The parser module implements the class ParseArgs which includes
all the methods required to parse any arguments entered either 
via the command line or through a list.

The class initialization checks for the existence of SAS_PATH.
Then it populates the pysaspkgs list with all packages below pysas.
This is used to get the version of the package.

The instance method optparser uses module argparse to define the
two types of arguments supported: options and parameters.
Options can be classified into two categories: immediate action and
execution modifiers. Options are accessed via the a single or a 
double '-'. All SAS tasks allow for a given set of specific options.
Parameters must include the '='symbol to separate the name and its
value. Within the value there can be more '=' symbols. 

Method procopt either executes the immediate options or sets the 
environment variables that modify the execution of the task.
"""


# Standard library imports
import os
import sys
import subprocess
import argparse
from importlib import import_module
import importlib.resources

# Third party imports

# Local application imports
from pysas.param import paramXmlInfoReader
from pysas.logger import get_logger
from pysas import sas_cfg

class ParseArgs:
    """
    For pySAS v2.0 it is now assumed that the input arguments are
    passed in as a dictionary.
    """
    def __init__(self, taskname, argdict, logger = None):
        self.taskname = taskname
        self.argdict = argdict
        if logger is None:
            # By default will only output to terminal
            self.logger = get_logger('ParseArgs')
        else:
            # Use logger that was passed in
            self.logger = logger

        # Before any further processing check that SAS_PATH is defined
        sas_path = os.environ.get('SAS_PATH')
        if not sas_path:
            raise Exception('SAS_PATH is undefined. SAS not initialised?')

        pysaspkgs = []

        my_resources = importlib.resources.files("pysas")
        for line in (my_resources / "pysaspkgs").read_text().splitlines():
            pysaspkgs.append(line)

        # If taskname is not in pysaspkgs it is a non Python SAS task
        # then its version must be obtained differently.
        # If it is a Python task, version is available from __version__ object
        if self.taskname in pysaspkgs:
            m = import_module('pysas.' + self.taskname + '.' + self.taskname)
            self.version = m.__version__
        else:
            # Do not use --version because some SAS perl tasks like epchain do not 
            # accept this option but only -v
            cmd = self.taskname + ' -v'
            self.version = subprocess.check_output(cmd, shell=True, text=True)

    # This is the task parser constructor
    def optparser(self):
        """
        This parses the 'options' from the input.
        """

        # Define the parser. No need to make it class wide.
        parser = argparse.ArgumentParser(
            prog = self.taskname,
            description = 'Parses command parameters and options',
            add_help = False
        )

        # Add the options with add_argument
        #
        # Two groups:
        # a) Immediate action:
        #    If the option is present, it must be processed immediately and exit.
        #    These are: -v/--version, -d/--dialog, -m/--manpage, -h/--help, -p/--param.

        parser.add_argument('-v', '--version', dest='version', action='store_true', default=False)
        parser.add_argument('-d', '--dialog', dest='dialog', action='store_true',default=False)
        parser.add_argument('-m', '--manpage', dest='manpage', action='store_true',default=False)
        parser.add_argument('-h', '--help', dest='help', action='store_true', default=False)
        parser.add_argument('-p', '--param', dest='param', action='store_true', default=False)

        # b) Modifiers:
        #    If the option is present, it modifies the
        #    execution of the command and/or the environment, usually
        #    by setting an environment variable.
        #    These are: -V/--verbosity (SAS_VERBOSITY), -c/--noclobber
        #    (SAS_CLOBBER), -a/--ccfpath (SAS_CCFPATH),
        #    -i/--ccf (SAS_CCF), -o/--odf (SAS_ODF), -f/--ccffiles,
        #    -w/--warning, -t/--trace.

        # Get default verbosity
        default_verbosity = sas_cfg.get_setting("verbosity")

        parser.add_argument('-V',
                            '--verbosity',
                            dest='verbosity',
                            type=int,
                            nargs=1,
                            default=default_verbosity,
                            choices=range(11),
                            action='store')

        parser.add_argument('-c',
                            '--noclobber',
                            dest='noclobber',
                            action='store_true',
                            default=False)

        parser.add_argument('-a',
                            '--ccfpath',
                            dest='sas_ccfpath',
                            type=str,
                            nargs=1,
                            action='store')

        parser.add_argument('-i',
                            '--ccf',
                            dest='sas_ccf',
                            action='store',
                            nargs=1)

        parser.add_argument('-o',
                            '--odf',
                            dest='sas_odf',
                            type=str,
                            nargs=1)

        parser.add_argument('-f',
                            '--ccffiles',
                            dest='ccffiles',
                            type=str,
                            nargs=1)

        parser.add_argument('-w',
                            '--warning',
                            dest='warning',
                            type=str,
                            nargs=1,
                            action='store')

        parser.add_argument('-t',
                            '--trace',
                            dest='trace',
                            action='store_true',
                            default=False)

        # Apply parse_args method to parser with the options being passed in. 
        # Result is put into parsedargs. The parsedargs object is not a 
        # dictionary but can be shown as such with function vars().

        self.parsedargs = parser.parse_args([self.argdict['options']])

        #print(self.parsedargs)
        #print(f'vars(self.parsedargs)          = \n{vars(self.parsedargs)}')
        #print(f'self.parsedargs.pvpairs        = {self.parsedargs.pvpairs}')

    # If these options are present, will execute a command and return Exit
    def exe_options(self):
        """
        Method to execute options or parameters that 
        require immediate action.

        These include options that execute the command 
        and exit. These are: 

            version(--version, -v)
            help(--help, -h)
            param(--param, -p)
            dialog(--dialog, -d)
            manpage(--manpage, -m)

        Returns 'Exit' which if True will send the exit 
        command up the chain. If False, then pySAS will 
        continue to execute.

        # Process options entered in command line
        #
        # Options version(--version, -v), help(--help, -h), param(--param, -p)
        # dialog(--dialog, -d) and manpage(--manpage, -m) are exclusive.  
        # Only one can be present and will set Exit to True. 
        # When exe_options is executed on the instance of ParseArgs (e.g. p), as
        # p.exe_options() the return Exit determines wheter to return immediately or
        # not. If Exit is set to True, the method will return immediately.
        #
        # Some options entered in self.argdict will launch special SAS tasks
        # designed only to provide specific result. These are:
        #
        # -d/--dialog   => sasdialog to launch task GUI
        # -m/--manpage  => sashelp to show HTML doc in the default web browser
        #

        """

        # Exit is set to False
        Exit = False

        # Define header text for option -h/--help
        usage = f"""
        Usage: {self.taskname} [Options] param0=value0 [param1=value1] ...

        Options:
        -a | --ccfpath <dir1>:<dir2>...   Sets SAS_CCFPATH to <dir1>:<dir2>...
        -c | --noclobber                  Set SAS_CLOBBER=0
        -d | --dialog                     Launchs task GUI
        -f | --ccffiles <f1> <f2> ...     CCF files
        -h | --help                       Shows this message, display param file contents and exits
        -i | --ccf <cif>                  Sets SAS_CCF=<cif> (ccf.cif)
        -m | --manpage                    Opens web browser with task documentation
        -o | --odf <sumfile>              Sets SAS_ODF to SAS summary file
        -p | --param                      List of all available task parameters with default values
        -t | --trace                      Trace task execution
        -V | --verbosity <level>          Sets verbosity level and sets SAS_VERBOSITY
        -v | --version                    Shows task name and version, and exits
        -w | --warning [code|n]           Set warning code to [code|n]

        Parameters:
        param0=value0                     A mandatory parameter. If not present, exits.
        [param1=value1]                   An optional parameter. If not present, it might have a default value
        ...
        """

        # Version:
        if self.parsedargs.version:
            print(f'{self.version}')
            Exit = True
        # Help:
        if self.parsedargs.help:
            print(usage)
            pf = paramXmlInfoReader(self.taskname, logger = self.logger)
            pf.xmlParser()
            pf.printHelp()
            Exit  = True
        if self.parsedargs.param:
            pf = paramXmlInfoReader(self.taskname, logger = self.logger)
            pf.xmlParser()
            pf.printHelp()
            Exit = True
        # dialog (GUI)
        if self.parsedargs.dialog:
            print(f'\nLaunching {self.taskname} GUI ...')
            cmd = ['sasdialog', self.taskname]
            self.runext(cmd)
            Exit = True
        # manpage (HTML doc)
        if self.parsedargs.manpage:
            print(f'\nLaunching web browser to display HTML documentation on {self.taskname} ...')
            cmd = ['sashelp', 'doc=' + self.taskname]
            self.runext(cmd)
            Exit = True

        if Exit:
            return Exit

    # If these options are present, they will be returned for handling
    def env_options(self):
        """
        Method to collect options that set environment variables 
        and then continue running the SAS task. These are:

               -V/--verbosity (SAS_VERBOSITY)
               -c/--noclobber (SAS_CLOBBER)
               -a/--ccfpath (SAS_CCFPATH)
               -i/--ccf (SAS_CCF)
               -o/--odf (SAS_ODF)
               -f/--ccffiles
               -w/--warning
               -t/--trace
        """
        # return_options is a dictionary to hold set options
        return_options = {}
        env_options_list = []

        # The options are accumulative, if they are 
        # present they are collected into a list and
        # returned to MyTask. 

        # sas_ccfpath
        if self.parsedargs.sas_ccfpath:
            envar_val = str(self.parsedargs.sas_ccfpath[0])
            env_options_list.append(f'-a {envar_val}')
            return_options['SAS_CCFPATH'] = envar_val
            self.logger.info(f'SAS_CCFPATH = {envar_val}')
        # noclobber
        if self.parsedargs.noclobber:
            envar_val = False
            env_options_list.append('-c')
            return_options['SAS_CLOBBER'] = envar_val
            self.logger.info(f'SAS_CLOBBER = {envar_val}')
        # sas_ccf
        if self.parsedargs.sas_ccf:
            envar_val = str(self.parsedargs.sas_ccf[0])
            env_options_list.append(f'-i {envar_val}')
            return_options['SAS_CCF'] = envar_val
            self.logger.info(f'SAS_CCF = {envar_val}')
        # sas_odf
        if self.parsedargs.sas_odf:
            envar_val = str(self.parsedargs.sas_odf[0])
            env_options_list.append(f'-o {envar_val}')
            return_options['SAS_ODF'] = envar_val
            self.logger.info(f'SAS_ODF = {envar_val}')
        # ccffiles
        if self.parsedargs.ccffiles:
            envar_val = str(self.parsedargs.sas_odf[0])
            env_options_list.append(f'-f {envar_val}')
            return_options['CCF_files'] = envar_val
            self.logger.info(f'CCF files = {envar_val}')
        # warning
        if self.parsedargs.warning:
            envar_val = str(self.parsedargs.warning[0])
            env_options_list.append(f'-w {envar_val}')
            return_options['WARNING'] = envar_val
            self.logger.info(f'WARNING = {envar_val}')
        # trace
        if self.parsedargs.trace:
            envar_val = True
            env_options_list.append('-t')
            return_options['TRACE'] = envar_val
            self.logger.info(f'Trace = {envar_val}')
        # verbosity
        if self.parsedargs.verbosity:
            envar_val = str(self.parsedargs.verbosity[0])
            env_options_list.append(f'-V {envar_val}')
            return_options['SAS_VERBOSITY'] = envar_val
            self.logger.info(f'SAS_VERBOSITY = {envar_val}')

        # Add the list of all options to return_options
        return_options['env_options'] = " ".join(env_options_list)

        return return_options

    def __repr__(self):
        return f'{self.__class__.__name__}({self.taskname} - {self.argdict})'

    def runext(self, runcmd):
        self.runcmd = runcmd
        if self.runcmd == None:
            return
        with subprocess.Popen(self.runcmd, bufsize=1, shell=False, text=True, stdout=subprocess.PIPE,stderr=subprocess.STDOUT, universal_newlines=True) as p:
            for line in p.stdout:
                print(line, end='')
        #if p.returncode != 0:
        #    raise subprocess.CalledProcessError(p.returncode, p.args)



#!/usr/bin/env python

#	ESA (C) 2000-2018 
#	This file is part of ESA's XMM-Newton Scientific Analysis System
#	(SAS).
#
#	SAS is free software: you can redistribute it and/or modify
#	it under the terms of the GNU General Public License as published by
#	the Free Software Foundation, either version 3 of the License, or
#	(at your option) any later version.
#
#	SAS is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#	   
#	You should have received a copy of the GNU General Public License
#	along with SAS.  If not, see <http://www.gnu.org/licenses/>.

import logging
from matplotlib.backends.backend_qt5agg import FigureCanvasQTAgg as FigureCanvas
from matplotlib.backends.backend_qt5agg import NavigationToolbar2QT as NavigationToolbar
from matplotlib.figure import Figure
import matplotlib.ticker as ticker
import matplotlib.pyplot as plt
import numpy as np
from astropy.io import fits
from astropy.wcs import WCS
from astropy.table import unique, Table
from PyQt5.QtWidgets import QSizePolicy
#import pylab as plt
from matplotlib.colors import LogNorm
import matplotlib.gridspec as gridspec
from matplotlib.ticker import  MaxNLocator, FormatStrFormatter, MultipleLocator, FuncFormatter, ScalarFormatter, NullFormatter,StrMethodFormatter

class PlotCanvas(FigureCanvas):
 
    def __init__(self, parent=None, width=648, height=648, dpi=100,fileName='test.ds',type=None,threshold=None,emllistFileName='test.ds',rgsEnerFileName='test.ds',ra=None, dec=None):
        #self.fig = Figure(figsize=(width, height), dpi=1)

        self.logger = logging.getLogger('xmmextractorGUI')
        self.logger.info("Creating Canvas")
        self.fig = Figure(figsize=(width, height), dpi=dpi,frameon=False)    

        self.fig.subplots_adjust(0.15,0.07,0.95,0.92,0.1,0.1)
        self.fileName = fileName
        self.emlistFileName = emllistFileName
        self.rgsEnerFileName = rgsEnerFileName
        self.lines = []
        self.plotCurve = []
        self.wcs = None
        self.ra = ra
        self.dec = dec
        FigureCanvas.__init__(self, self.fig)
        self.setParent(parent)
        
        #FigureCanvas.setSizePolicy(self,
        #        QSizePolicy.Expanding,
        #        QSizePolicy.Expanding)
        #FigureCanvas.updateGeometry(self)
        #self.fig.frameon = False

        self.fig.set_size_inches((280,60))
        if type == 'IM':
            self.plot()
        elif type == 'LC':
            self.plotLC(threshold,self.fileName)
        elif type == 'PG':
            self.plotPG(threshold,self.fileName)            
        elif type == 'DT':
            self.plotSrcDetection()
        elif type == 'RGS':
            self.plotRGS()
        elif type == 'RGSFlux':
            self.plotRGSFlux()
        elif type == 'RGSLC':
            self.plotRGSLC()
        else:
            self.plotSP()
        


    def plotSrcDetection(self):

        self.logger.info("plotting image")
        if self.fileName != "NOT FOUND":

            #Read the source detection list 
            table = Table.read(self.emlistFileName, hdu=1)  

            uTable = unique(table, keys='ML_ID_SRC')
                     
            image_data,hdu = fits.getdata(self.fileName, ext=0, header=True)
            self.wcs = WCS(hdu)
            #print(image_data.shape)

            ax = self.fig.add_subplot(111,projection=self.wcs)
            ra = ax.coords[0]
            dec = ax.coords[1]
            ra.set_major_formatter('d.ddd')
            dec.set_major_formatter('d.ddd')

            ax.clear()
            ax.imshow(image_data,cmap='hot', norm=LogNorm())  
 
            ax.set_title('Full Image')
            ax.set_xlabel("RA")
            ax.set_ylabel("DEC")


            for row in uTable:                
                ax.add_artist(plt.Circle((row['X_IMA'], row['Y_IMA']), 5.5, color='red',lw=2, fill=False))

            self.fig.canvas.mpl_connect('button_press_event',self.onclick)

            self.toolbar = NavigationToolbar(self.fig.canvas, self,coordinates=False)
            self.toolbar.setMinimumWidth(300)

            
            #This call takes lot of time...
            #self.draw()
         

    def plot(self):

        self.logger.info("plotting image")
        if self.fileName != "NOT FOUND":

            ax = None
            image_data,hdu = fits.getdata(self.fileName, ext=0, header=True)
            if hdu['DATAMODE'] != "TIMING":
                self.wcs = WCS(hdu)
                ax = self.fig.add_subplot(111,projection=self.wcs)
            else:
                ax = self.fig.add_subplot(111)
                            
            if hdu['DATAMODE'] != "TIMING":
                ra = ax.coords[0]
                dec = ax.coords[1]
                ra.set_major_formatter('d.ddd')
                dec.set_major_formatter('d.ddd')

            ax.clear()
            ax.imshow(image_data,cmap='hot', norm=LogNorm())  
          

            if self.ra is not None:
                if hdu['DATAMODE'] != "TIMING":               
                    px, py = self.wcs.wcs_world2pix(self.ra,self.dec, 1)
                    ax.add_artist(plt.Circle((px, py), 5.5, color='white',lw=2, fill=False))

            ax.set_title('Full Image')

            if hdu['DATAMODE'] != "TIMING":
                ax.set_xlabel("RA")
                ax.set_ylabel("DEC")                
                self.fig.canvas.mpl_connect('button_press_event',self.onclick)

            self.toolbar = NavigationToolbar(self.fig.canvas, self,coordinates=False)
            self.toolbar.setMinimumWidth(300)
            #This call takes lot of time...
            #self.draw()
        
    def format_fn(tick_val, tick_pos):
            labels = list('abcdefghijklmnopqrstuvwxyz')
            if int(tick_val) in xs:
                return labels[int(tick_val)]
            else:
                return ''

    def major_formatter(x, pos):
            return "[%.2f]" % x

    def plotRGS(self):
        self.logger.info("plotting image")
        if self.fileName != "NOT FOUND" or self.rgsEnerFileName != "NOT FOUND": 

            image_data,hdu = fits.getdata(self.fileName, ext=0, header=True)
            ener_image_data,hdu2 = fits.getdata(self.rgsEnerFileName, ext=0, header=True)
            gs = gridspec.GridSpec(2, 1,height_ratios=[1, 1],width_ratios=[1])
            ax1= self.fig.add_subplot(gs[0])
            #ax1 = self.fig.add_subplot(221)
            ax1.clear()
            ax1.imshow(image_data,cmap='hot', norm=LogNorm(),origin='lower',aspect='auto')  
            text = hdu['INSTRUME']+" Spatial and Orders Image"
            ax1.set_title(text)
            #ax1.set_xlabel("LAMBDA")
            ax1.set_ylabel("Cross Dispersion Angle (radians)")
            ax2= self.fig.add_subplot(gs[1],sharex=ax1)
            #ax2 = self.fig.add_subplot(121)
            ax2.clear()

            
            ax2.imshow(ener_image_data,cmap='hot', norm=LogNorm(),origin='lower',aspect='auto')  
            ax2.set_xlabel("LAMBDA")

            crval1 = hdu2['CRVAL1']
            delta1 = hdu2['CDELT1']
            val = crval1

            xticks = range(0, len( ener_image_data[1]), 1)
            labels=[0]*int(len( ener_image_data[1])/1)
            counter = 0
            for lTick in xticks:
                labels[counter]=  ("%.1f" % (crval1+(lTick*delta1)))
                counter = counter + 1
           
            #ax2.xaxis.set_major_locator(MultipleLocator(50.00))     
                
            ax2.xaxis.set_major_formatter(FuncFormatter(lambda x,pos: "%.2f" % (crval1+(x*delta1))))
            #ax2.set_xticklabels(labels,minor=False)
            #ax2.set_xticks(xticks)
            ax2.set_ylabel("PI (Channel)")
                
            self.fig.canvas.mpl_connect('button_press_event',self.onclickRGS)

            self.toolbar = NavigationToolbar(self.fig.canvas, self,coordinates=False)
            self.toolbar.setMinimumWidth(300)
            #This call takes lot of time...
            #self.draw()
            
        
    def onclick(self,event):        
        if event.xdata is not None and event.ydata is not None:
            print('x=%d, y=%d, xdata=%f, ydata=%f' %
                   (event.x, event.y, event.xdata, event.ydata))
            wx,wy=self.wcs.wcs_pix2world(event.xdata,event.ydata,1)
            print('RA: {0} DEC: {1}'.format(wx,wy))

    def onclickRGS(self,event):
        if event.xdata is not None and event.ydata is not None:
            print('x=%d, y=%d, xdata=%f, ydata=%f' %
                  (event.x, event.y, event.xdata, event.ydata))
        
        
 
    def plotLC(self,threshold,fileName):        
        self.logger.info("plotting lightcurve") 
        self.fileName = fileName
        if self.fileName != "NOT FOUND":
            fitsFile = fits.open(self.fileName)   
            prihdu = fitsFile[1].header
            if ('CUTVAL' in prihdu):
                threshold = prihdu['CUTVAL']

            cols = fitsFile[1].columns
            colName = None
            for i,x  in enumerate(cols.names):
                if "RATE" in x:
                    colName = cols.names[i]
                if "COUNTS" in x:
                    colName = cols.names[i]        

            data = fitsFile[1].data   
            
            xdata = data.field('TIME')           # extract the x data column
            ydata = data.field(colName)

            ax = self.fig.add_subplot(111)
            ax.clear()
            start, end = ax.get_xlim()
            ax.xaxis.get_major_formatter().set_useOffset(False)
            ax.plot(xdata,ydata)               # plot the data

            if colName == 'RATE' and "gti.fit" not in self.fileName :
                ax.set_title("Background substracted lightcurve")
                ax.set_xlabel("Time (s)")
                ax.set_ylabel("Cts/s")
            else:
                ax.set_title("High-particle flaring lightcurve")
                ax.set_xlabel("Time (s)")
                ax.set_ylabel("Counts")

            
            if (threshold != 'None'):
                if (colName == 'COUNTS'):
                    threshold=float(threshold)*100.

                y2data = [threshold]*len(xdata)            
                ax.plot(xdata,y2data)
            
            self.toolbar = NavigationToolbar(self.fig.canvas, self,coordinates=False)
            self.toolbar.setMinimumWidth(300)
            
            fitsFile.close()
            
        #self.draw()

    def plotPG(self,threshold,fileName):
        self.logger.info("plotting S/N curve")
        self.fileName = fileName

        if self.fileName != "NOT FOUND":

            fitsFile = fits.open(self.fileName)   
            data = fitsFile[1].data   
            
            xdata = data.field('BKG')           # extract the x data column
            ydata = data.field('SN')

            ax = self.fig.add_subplot(111)
            #ax.clear()
            #Remove current plot 
            for i, line in enumerate(self.plotCurve):                
                self.plotCurve.pop(i)
                line.remove()
            
            #Remove exisiting lines (if any)            
            for i, line in enumerate(self.lines):
                self.lines.pop(i)
                line.remove()
                
            self.plotCurve = ax.plot(xdata,ydata)               # plot the data

            if threshold == 'None':
                ax.set_title("Background substracted lightcurve")
                ax.set_xlabel("Time (s)")
                ax.set_ylabel("Cts/s")
            else:
                ax.set_title("High-particle flaring lightcurve")
                ax.set_xlabel("BKG Counts")
                ax.set_ylabel("S/N")
                x2data = [threshold]*len(xdata)
                self.lines = ax.plot(x2data,ydata)
            

            self.toolbar = NavigationToolbar(self.fig.canvas, self,coordinates=False)
            self.toolbar.setMinimumWidth(300)
            fitsFile.close()
            #self.draw()
   
    def plotSP(self):
        self.logger.info("plotting spectra")
        if self.fileName != "NOT FOUND":
            fitsFile = fits.open(self.fileName)   
            data = fitsFile[1].data   
            
            xdata = data.field('CHANNEL')           # extract the x data column
            ydata = data.field('COUNTS')

            ax = self.fig.add_subplot(111)
            ax.clear()
            ax.plot(xdata,ydata)               # plot the data
            ax.set_title("Spectrum")
            ax.set_xlabel("Channel")
            ax.set_ylabel("Counts")

            self.toolbar = NavigationToolbar(self.fig.canvas, self,coordinates=False)
            self.toolbar.setMinimumWidth(300)
            fitsFile.close()

    def plotRGSFlux(self):
        self.logger.info("plotting RGS Flux spectra " +self.fileName )
        if self.fileName != "NOT FOUND":
            fitsFile = fits.open(self.fileName)   
            data = fitsFile[1].data   
            
            xdata = data.field('CHANNEL')           # extract the x data column
            ydata = data.field('FLUX')

            ax = self.fig.add_subplot(111)
            ax.clear()
            ax.plot(xdata,ydata)               # plot the data
            ax.set_title("Fluxed Spectrum")
            ax.set_xlabel("Channel")
            ax.set_ylabel("Flux")

            self.toolbar = NavigationToolbar(self.fig.canvas, self,coordinates=False)
            self.toolbar.setMinimumWidth(300)
            fitsFile.close()


    def plotRGSLC(self):
        self.logger.info("plotting RGS lightcurve") 

        if self.fileName != "NOT FOUND":
            fitsFile = fits.open(self.fileName)   
            prihdu = fitsFile[1].header
            instrument = prihdu['INSTRUME']
            data = fitsFile[1].data   
            
            xdata = data.field('TIME')           # extract the x data column
            ydata = data.field('RATE')

            ax = self.fig.add_subplot(111)
            ax.clear()
            start, end = ax.get_xlim()
            ax.xaxis.get_major_formatter().set_useOffset(False)
            ax.plot(xdata,ydata)               # plot the data

            title = instrument +" Background substracted lightcurve"
            ax.set_title(title)
            ax.set_xlabel("Time (s)")
            ax.set_ylabel("Cts/s")
            
            self.toolbar = NavigationToolbar(self.fig.canvas, self,coordinates=False)
            self.toolbar.setMinimumWidth(300)
            
            fitsFile.close()
            



# ESA (C) 2000-2021
# 
# This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
#
#    SAS is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    SAS is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
#


# Standard library imports
import copy
import logging
import os
import sys
from pathlib import Path

# Third party imports
from loguru import logger
# This cleans up any loguru loggers that
# might be hanging around.
logger.remove()

# Local application imports
from pysas import sas_cfg

# Functions

def get_logger(taskname: str, 
               toterminal  = True, 
               tofile      = False, 
               logfilename = None, 
               tasklogdir  = None,
               pylogger    = True):
    """
    Function to get a loguru logger object.

    Inputs:
    (required)
        - taskname   : Name of the task to be run. By default will
                       name log file "{taskname}.log"
    (optional)
        - toterminal : (default: True) Output will be written to 
                        the terminal.
        - tofile     : (default: False) Output will be written to 
                        a log file.
        - logfilename: Designated log file name. Will be used instead
                       of "{taskname}.log". Useful for putting all 
                       output from multiple tasks into the same file.
        - tasklogdir : (default: cwd) Directory where to write the log 
                       file.
                       Priority of defaults for task_logdir
                         1. tasklogdir (passed in to function)
                         2. SAS_TASKLOGDIR (envirnment variable)
                         3. cwd (final default)
        - pylogger   : True: Use logger settings for Python logging
                       False: Use logger settings for subprocess logging
                       Only set to 'False' in sastask.py -> MyTask.run()
                       when running a non Python SAS task using 
                       subprocess.
    """
    task_logger = copy.deepcopy(logger)
    
    # SAS_TASKLOGDIR allows to set the directory for the logging file
    # Priority of defaults for task_logdir
    #   1. tasklogdir (passed in to function)
    #   2. SAS_TASKLOGDIR (envirnment variable)
    #   3. cwd (final default)
    sas_tasklogdir = os.getenv('SAS_TASKLOGDIR')

    if (tasklogdir and os.path.isdir(tasklogdir)):
        task_logdir = Path(tasklogdir)
    else:
        if(sas_tasklogdir and os.path.isdir(sas_tasklogdir)):
            task_logdir = Path(sas_tasklogdir)
        else:
            task_logdir = Path.cwd()

    if logfilename:
        task_logfile = task_logdir / logfilename
    else:
        task_logfile = task_logdir / f"{taskname}.log"

    # SAS_TASKLOGFMODE allows to set the write mode for the logging file
    # Allowed modes are : w (new file each invocation of logger),
    # a (append to any existing file)A That is the default mode.
    sastasklogfmode = os.getenv("SAS_TASKLOGFMODE", "a")
    if sastasklogfmode != 'a' and sastasklogfmode != 'w':
        sastasklogfmode = 'a'

    if pylogger:
        # For Python based logging
        # Set verbosity
        verbosity = int(os.getenv('SAS_VERBOSITY'))
        match(verbosity):
            case 1:
                level = "CRITICAL"
            case 2 | 3:
                level = "ERROR"
            case 4 | 5:
                level = "WARNING"
            case 6 | 7:
                level = "INFO"
            case 8 | 9 | 10:
                level = "DEBUG"
            case _:
                level = "INFO"
        
        # 98% of the time pysas_verbosity == SAS_VERBOSITY.
        # But for debugging purposes it is convienent to have
        # a separate verbosity for pySAS.
        # This does not affect the verbosity for non-Python
        # based SAS tasks.
        pysas_verbosity = sas_cfg.get_setting('pysas_verbosity')

        # Take the lower of sas_level or pysas_level
        sas_level   = logger.level(level).no
        pysas_level = logger.level(pysas_verbosity).no
        if sas_level < pysas_level: pysas_verbosity = level

        level = pysas_verbosity
        
        # Add file sink
        if tofile:
            task_logger.add(
                sink=task_logfile,
                level=level,
                mode=sastasklogfmode,
                enqueue=True,
                format="{time:DD-MM-YYYY HH:mm:ss.SSS Z} - {name} - {level: <8} - {message}"
            )

        # Add console sink
        if toterminal:
            task_logger.add(
                sink=sys.stdout,
                level=level,
                enqueue=True,
                format="<g>{name}</g> - <level>{level: <8}</level> - <level>{message}</level>"
            )
    else:
        # For non-Python based logging (i.e. subprocess)
        # Set verbosity
        # For non-Python SAS tasks the verbosity is set separately
        level = "INFO"
        
        # Add file sink
        if tofile:
            task_logger.add(
                sink=task_logfile,
                level=level,
                mode=sastasklogfmode,
                enqueue=True,
                format="<green>{time:DD-MM-YYYY HH:mm:ss.SSS Z}</green> - {message}"
            )

        # Add console sink
        if toterminal:
            task_logger.add(
                sink=sys.stdout,
                level=level,
                enqueue=True,
                format="{message}",
            )

    return task_logger

class TaskLogger:
    """
    TaskLogger creates a logger for Python tasks.
    The name of the logger is the name of the task.
    Such logger is known as saslogger. 
    It includes two handlers, one console handler
    and one file handler. The file handler will put
    all logging messages in a file named <taskname>.log.
    
    This file is by default written in the starting directory
    of the task (os.getcwd()). However, for each task we can 
    set that directory by taking into account the env variable 
    SAS_TASKLOGDIR, if defined. 
    
    By default the logging file is created in  mode -a- (append)
    to allow to use the same file to log several runs of the
    same task. However, the SAS_TASKLOGFMODE environment variable
    can be set to -w- (new file), to change that behaviour.

    File handler has alogging level of DEBUG, so everything
    is logged. However, console handler logging level is 
    controlled by the SAS Verbosity (SAS_VERBOSITY env. variable)
    which is set via the option -V.

    The logger can not be called more than once. If not, 
    messages start appearing duplicated. That is why the
    logging method is separarted from the logger definition.
    """

    def __init__(self, taskname):
        self.taskname = taskname
       
        self.dictlevels = {
                'notset':0,
                'debug': 10,
                'info': 20,
                'warning': 30,
                'error': 40,
                'critical': 50
                }

        self.saslogger = logging.getLogger(self.taskname)
        self.saslogger.setLevel('DEBUG')

        # SAS_TASKLOGDIR allows to set the directory for tyhe logging file
        sastasklogdir = os.getenv('SAS_TASKLOGDIR')
        startdir = os.getcwd()
        #print(f'{self.taskname} was initiated from {startdir}')
       
        if not sastasklogdir:
            #print(f'\nTask logging directory is unset. \nUsing {startdir} instead.')
            sastasklogdir = startdir
        if not os.path.isdir(sastasklogdir):
            #print(f'\n{sastasklogdir} does not exist. \nUsing {startdir} instead.')
            sastasklogdir = startdir

        self.logfile = os.path.join(sastasklogdir, self.taskname + '.log')
        #print(f'\n{self.taskname} log file: {self.logfile}\n')

        # Clear any previous handlers
        if self.saslogger.hasHandlers:
            self.saslogger.handlers.clear()

        # SAS_TASKLOGFMODE allows to set the write mode for the logging file
        # Allowed modes are : w (new file each invocation of logger),
        # a (append to any existing file)A That is the default mode.

        sastasklogfmode = os.getenv('SAS_TASKLOGFMODE')
        if not sastasklogfmode:
            sastasklogfmode = 'a'
        if sastasklogfmode != 'a' and sastasklogfmode != 'w':
            sastasklogfmode = 'a'
        #print(f'\nTask logging file mode = {sastasklogfmode}')

        # File handler (fh). Level fixed to DEBUG
        self.fh = logging.FileHandler(self.logfile, mode=sastasklogfmode)
        self.fh.setLevel('DEBUG')

        # Console handler (ch). Default level set to WARNING
        self.ch = logging.StreamHandler()
        self.ch.setLevel('WARNING')

        # Add handlers to the logger
        self.saslogger.addHandler(self.ch)
        self.saslogger.addHandler(self.fh)

        # Create formatters and add them to handlers
        cformat = logging.Formatter('%(name)s - %(levelname)s - %(message)s')
        fformat = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s', datefmt='%d-%b-%y %H:%M:%S')

        self.ch.setFormatter(cformat)
        self.fh.setFormatter(fformat)


    def __repr__(self):
        return f'{self.__class__.__name__}({self.taskname})'


    def log(self, level, msg):
        """
        A log message will be issued according to the logging level.
        The logging level is fixed for the log file but 
        on the console will depend on SAS_VERBOSITY.
        SAS_VERBOSITY can be set between 1 (min) and 10 (max).
        Correspondence between SAS_VERBOSITY levels and logging levels
        DEBUG      ->   verbosity  = 8, 9, 10
        INFO       ->   verbosity  = 6, 7
        WARNING    ->   verbosity  = 4, 5
        ERROR      ->   verbosity  = 2, 3
        CRITICAL   ->   verbosity  = 1
        """

        v = os.getenv('SAS_VERBOSITY')
        if v == None:
            v = '4'
        verbosity = int(v)

        if   verbosity >= 8 and verbosity <= 10:
            self.ch.setLevel('DEBUG')
        elif verbosity >= 6 and verbosity <= 7:
            self.ch.setLevel('INFO')
        elif verbosity >= 4 and verbosity <= 5:
            self.ch.setLevel('WARNING')
        elif verbosity >= 2 and verbosity <= 3:
            self.ch.setLevel('ERROR')
        elif verbosity == 1:
            self.ch.setLevel('CRITICAL')

        nlevel = self.dictlevels[level]
        self.saslogger.log(nlevel, msg)




# ESA (C) 2000-2020
# 
# This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
#
#    SAS is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    SAS is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with SAS.  If not, see <http://www.gnu.org/licenses/>.

# pyomgchain.py

from .version import VERSION
from pysas import SAS_RELEASE, SAS_AKA

__version__ = f'pyomgchain (pyomgchain-{VERSION}) [{SAS_RELEASE}-{SAS_AKA}]' 

# The Python code for pyomgchain should go inside run

# iparsdic is a dictionary with all the task parameters, where their
# respective values are either those entered from the command line 
# or those taken from par file defaults.

import os
import datetime
import sys
import glob
import pysas.pyutils.pyutils as pyutils
import logging
import subprocess
import re
import fnmatch
import time



def process_orbit(orbit_key, has_ff, ff_name):
    """
    Process the orbit.
    
    Args:
        orbit_key: the orbit to process.
        has_ff: the flat field flag.
        ff_name: the name of the flat field flag.
    """

    highlighted_message('*', '    Processing orbit {}'.format(orbit_key))

    global flat_field_filename
    for obs in obs_list.keys():
        Message('     Observation {}'.format(obs))
        if has_ff:
            Message('Using flatfield (hasff = {})'.format(has_ff))
            Message('ff_name = {}'.format(ff_name))
            flat_field_filename = ff_name
        else:
            flat_field_filename = ''

        for exp_key in sorted(exp_list.keys()):
            process_exposure(orbit_key, obs, exp_key)


def process_exposure(orb_key, obs_key, exp_key):
    """
    Processes a single exposure.

    Args:

        orb_key: orbit key
        obs_key: the observation key
        exp_key: the exposure.
    """

    vis_win = dict()
    vis_wx0 = dict()
    vis_wy0 = dict()
    vis_wdx = dict()
    vis_wdy = dict()
    vis_proc = dict()

    uv_win = dict()
    uv_wx0 = dict()
    uv_wy0 = dict()
    uv_wdx = dict()
    uv_wdy = dict()
    uv_proc = dict()

    message = '     Exposure: {}'.format(exp_key)
    highlighted_message('*', message)

    tracking_history_flag = 0
    window_file_flag =0

    if fnmatch.filter(list_of_files, orb_key + obs_key + exp_key + '*00THX*'):
        tracking_history_flag = 1
        tracking_history_filename = fnmatch.filter(list_of_files, orb_key + obs_key + exp_key + '*00THX*')[0]
    else:
        tracking_history_flag = 0

    if tracking_history_flag:
        Message('Tracking hisotry flag: Ok')
    else:
        Message('Tracking history filename file for {} is missing.'.format(exp_key))

        SAS_dir = os.environ['SAS_DIR']
        Message('Creating a dummy file for tracking history.')
        tracking_history_filename = 'DUMMYTHX.FIT'

    global window_data_filename
    if fnmatch.filter(list_of_files, orb_key + obs_key + exp_key + '*00WDX*'):
        window_file_flag = 1
        window_data_filename = fnmatch.filter(list_of_files, orb_key + obs_key + exp_key + '*00WDX*')[0]
    else:
        window_data_flag = 0
    if window_file_flag:
        Message('Window data file: o.k')
    else:
        Message('Window data file for the exposure {} is missing.'.format(exp_key))

    f1 = '*' + orb_key + '*' + obs_key + '*' + exp_key + '*00WDX*'
    f2 = '*' + orb_key + '*' + obs_key + '*OM*00000NPH*'
    f3 = '*' + orb_key + '*' + obs_key + '*OM*00000PEH*'

    if fnmatch.filter(list_of_files, f1) and fnmatch.filter(list_of_files, f2) and fnmatch.filter(list_of_files, f3):
        global periodic_hk_filename, non_periodic_hk_filename
        periodic_hk_filename = fnmatch.filter(list_of_files, f3)[0]
        non_periodic_hk_filename = fnmatch.filter(list_of_files, f2)[0]
        window_data_filename = fnmatch.filter(list_of_files, f1)[0]

        s = window_data_filename[18] # takes the letter for the mode
        tracking_history_plot_filename = 'P' + obs_key + 'OM' + s + exp_key + 'TSHPLT' + '0000.PS'
        Message('Tracking history plot file: {}'.format(tracking_history_plot_filename))

        s = non_periodic_hk_filename[18] # takes the letter for the mode
        tracking_history_timeseries_filename = 'P' + obs_key + 'OM' + s + exp_key + 'TSTRTS0000.FIT'
        Message('Tracking history timeseries file: {}'.format(tracking_history_timeseries_filename))

        number_of_grisms = 0
        vis_index = 0
        uv_index = 0

        for i in sorted(win_list.keys()):
            win_key = i
            if fnmatch.filter(list_of_files, orb_key + '*' + obs_key + '*' + exp_key + '*' + win_key + '*IMI*')[0]:
                image_file = inp_dir + '/' + fnmatch.filter(list_of_files, orb_key + '*' + obs_key + '*' + exp_key + '*' + win_key + '*IMI*')[0]
                if os.path.isfile(image_file):
                    Message('Image found: {}'.format(image_file))

                filter_id = pyutils.get_key_word(image_file, 'FILTER')
                wx0 = pyutils.get_key_word(image_file, 'WINDOWX0')
                wy0 = pyutils.get_key_word(image_file, 'WINDOWY0')
                wdx = pyutils.get_key_word(image_file, 'WINDOWDX')
                wdy = pyutils.get_key_word(image_file, 'WINDOWDY')

                if filter_id == 200:
                    vis_win.update({vis_index : win_key})
                    vis_wx0.update({vis_index : wx0})
                    vis_wy0.update({vis_index : wy0})
                    vis_wdx.update({vis_index : wdx})
                    vis_wdy.update({vis_index : wdy})
                    vis_proc.update({vis_index : 1})
                    vis_index = vis_index + 1

                if filter_id == 1000:
                    uv_win.update({uv_index : win_key})
                    uv_wx0.update({uv_index : wx0})
                    uv_wy0.update({uv_index : wy0})
                    uv_wdx.update({uv_index : wdx})
                    uv_wdy.update({uv_index : wdy})
                    uv_proc.update({uv_index : 1})
                    uv_index = uv_index + 1

        if vis_index > 0:
            process_windows(orb_key, obs_key, exp_key, filter_id, vis_index, vis_win, vis_wx0, vis_wy0, vis_wdx, vis_wdy, vis_proc)

        if uv_index > 0:
            process_windows(orb_key, obs_key, exp_key, filter_id, uv_index, uv_win, uv_wx0, uv_wy0, uv_wdx, uv_wdy, uv_proc)

    else:
        Message('*** Warning ***')
        Message('Exposure {} in observation {} is missing a file'.format(exp_key, obs_key))

        thx_file = orb_key + "_" + obs_key + "_OMS" + exp_key + "00THX.FIT"
        Message('Please check if the file {} is in the ODF'.format(thx_file))

        nph_file = orb_key + "_" + obs_key + "_OMS" + exp_key + "00NPH.FIT"
        Message('Please check if the file {} is in the ODF'.format(nph_file))

        peh_file = orb_key + "_" + obs_key + "_OMS" + exp_key + "00PEH.FIT"
        Message('Please check if the file {} is in the ODF'.format(peh_file))

        wdx_file = orb_key + "_" + obs_key + "_OMS" + exp_key + "00WDX.FIT"
        Message('Please check if the file {} is in the ODF'.format(wdx_file))


def Message(message):
    """
    Prints a message directly to the screen and sends it to a log file.
    
    Args:
        message: the message to send to the log and screen.
    """

    logging.basicConfig(filename = 'pyomgchain.log', level = logging.DEBUG)

    print(message)
    logging.info(message)


def highlighted_message(character, message):
    """
    Delivers a clearier, easier to read, and probably more important
    message.

    Args
        character: character preceding the message used to 'highlight'.
        message: the proper message.
    """

    message = message.split('\n')
    num_lines = len(message)
    max_lenght = len(message[0])

    if num_lines > 1:
        for i in range(1, num_lines):
            if len(message[i]) > max_lenght:
                max_lenght == len(message[i])

    text = ''
    for i in range(0, max_lenght):
        text = text + character
    Message(text)

    for i in range(0, num_lines):
        Message(message[i])

    Message(text)


def get_ODF_directory(SASfile):
    """
    Returns the ODF directory.

    Args:
        SAS_file: the .SAS summary file.
    
    Output: 
        directory: the ODF directory as a string.
    """

    with open (SASfile, 'r') as SAS_file:
        for line in SAS_file:
            if 'PATH' in line:
                directory = line[5:]

    return directory


def check_for_house_keeping_files(list_of_files):
    """
    Raises exceptions if there are missing periodic or non periodic files.

    Args:
        list_of_files: the total list of the files from the input directory.
    
    Output:
        (periodic_hk_filename, non_periodic_hk_filename): the house keeping files as a tuple of strings.
    """

    print('Checking for house keeping files...')
    periodic_file_flag = 0
    non_periodic_file_flag = 0

    for file in list_of_files:
        if file.endswith('PEH.FIT') and 'OM' in file:
            periodic_hk_filename  = file
            periodic_file_flag = 1
        if file.endswith('NPH.FIT') and 'OM' in file:
            non_periodic_file_flag = 1
            non_periodic_hk_filename = file

    if periodic_file_flag == 0:
        message = "The ODF does NOT contain a Periodic Housekeeping file. Please insert into the ODF."
        Message(message)
        raise OSError('Task Failure. Incomplete ODF.')
    else:
        Message(    "ODF contains a Periodic Housekeeping file.")

    if non_periodic_file_flag == 0:
        message = "The ODF does NOT contain a Non Periodic Housekeeping file. Please insert into the ODF."
        Message(message)
        raise OSError('Task Failure. Incomplete ODF.')
    else:
        Message('ODF contains a non periodic housekeeping file.')

    return (periodic_hk_filename, non_periodic_hk_filename)


def set_up_directory_paths():
    """
    Prepares some basic directories into global variables, if not already present.
    
    
    Output:
        (has_ff, ff_name, inp_dir, out_dir): tuple containing the flat field flag and filename and the paths of the input and output directories.
    """

    global inp_dir, out_dir
    has_ff = 0
    ff_dir = os.environ['SAS_DIR'] + '/packages'
    working_dir = os.getcwd()
    if os.path.isdir(ff_dir):
        ff_name = ff_dir + "/omgchain/test/superflat_withoutLED.fit"
        if os.path.isdir(ff_name):
            print('Flat field file found.')
            has_ff = 1
    else:
        ff_name = ''

    if inpdirectory == '':
        if not 'SAS_ODF' in os.environ or os.environ['SAS_ODF'] == '':
            inp_dir = os.getcwd()
            os.environ['SAS_ODF'] = inp_dir
            SAS_file = glob.glob('*.SAS')[0]

        else:
            pos = os.environ['SAS_ODF'].index('SAS')
            if pos > 0:
                SAS_file = os.environ['SAS_ODF']
                inp_dir = get_ODF_directory(SAS_file).rstrip()
            else:
                inp_dir = os.environ['SAS_ODF'].rstrip()
                SAS_file = glob.glob(inp_dir + '*.SAS')[0]
    else:
        if os.path.isdir(inpdirectory):
            inp_dir = inpdirectory
        else:
            os.mkdir(inpdirectory)
            inp_dir = inpdirectory

    if outdirectory != '':
        if os.path.isdir(outdirectory):
            out_dir = outdirectory
        else:
            os.mkdir(outdirectory)
            out_dir = outdirectory
    else:
        out_dir = os.getcwd()


    return(has_ff, ff_name, inp_dir, out_dir)


def process_windows(orb_key, obs_key, exp_key, filter_id, index, mixed_win, mixed_wx0, mixed_wy0, mixed_wdx, mixed_wdy, mixed_proc):
    """
    Don't mistake this with process_window. Evaluates the windows that are in each exposure.
    Args:
        orb_key: current orbital key.
        obs_key: current orbit key.
        exp_key: current exposure key.
        filter_id: the id of the filter for the exposure.'
        index: value for index taken from the exposure.
        mixed_win: the mixed value for the window.
        mixed_wx0, mixed_wy0, mixed_wdx, mixed_wdy, mixed_proc: mixed values taken from the fits from the exposure.
    """

    win = mixed_win
    wx0 = mixed_wx0
    wy0 = mixed_wy0
    wdx = mixed_wdx
    wdy = mixed_wdy
    proc = mixed_proc
    comb_list = dict()

    message_filter = 'filter value does not correspond to a grism'

    if filter_id == 200:
        grism_id = "Grism 2 (visual)"
        message_filter = "Filter value corresponds to {}".format(grism_id)

    if filter_id == 1000:
        grism_id = "Grism 1 (UV)"
        message_filter = "Filter value corresponds to {}".format(grism_id)

    Message(message_filter)

    i = 0
    while i < index:
        if proc[i] >= 0:
            wx_edge = wx0[i]
            dx_edge = wdx[i]
            wy_edge = wy0[i]
            dy_edge = wdy[i]
            message_window = "Window {} will be processed".format(win[i])
            Message(message_window)

            icomb = 0
            comb_list = list(win.values())

            # Combine images only if the parameter $combine is yes
            if combine:
                j = i
                Message('Combine = yes')
                while j < (index - 1):
                    j = j + 1
                    if wx0[j] == (wx_edge + dx_edge) and wy0[j] == wy_edge and wdy[j] == dy_edge:
                        icomb = icomb + 1
                        comb_list[icomb] = win[j]
                        wx_edge = wx_edge + dx_edge
                        dx_edge = wdx[j]
                        wy_edge = wy0[j]
                        dy_edge = wdy[j]
                        proc[j] = -1

                k = 0
                # Check if the input files exist:
                while k <= icomb:
                    f = orb_key + '*' + obs_key + '*' + exp_key + '*' + comb_list[k] + "*IMI*"
                    if fnmatch.filter(list_of_files, f):
                        input_image_filename = fnmatch.filter(list_of_files, f)[0]
                    k = k + 1

            k = len(comb_list)
            process_window(orb_key, obs_key, exp_key, comb_list)

        else:
            Message("The window {} is in the combined list and will not be processed ".format(win[i]))
        i = i + 1


def process_window(orb_key, obs_key, exp_key, comb_list):
    """
    Processes a single window of the observation.

    Args:
        orb_key: the current orbital key.
        obs_key: the current observation key.
        exp_key: the current exposition key.
        comb_list: the combination list for windoes.
    """

    exposure_image_list = []
    n_win = len(comb_list)
    combined_file_name = "g" + obs_key + "OMS" + exp_key + "CIMAGE0000.FIT"
    mode_symbol = 'S'
    combined_image_was_produced = 0

    Message('Number of sub-windows found: {}'.format(n_win))

    if combine and n_win == 4:
        k = 0
        while k <= n_win:
            f = orb_key + '*' + obs_key + '*' + exp_key + '*' + comb_list[k] + "*IMI*"
            if fnmatch.filter(list_of_files, f):
                exposure_image_list.append(inp_dir + '/' + fnmatch.filter(list_of_files, orb_key + '*' + obs_key + '*' + exp_key + '*' + comb_list[k] + '*IMI*')[0])
                mode_symbol = fnmatch.filter(list_of_files, orb_key + '*' + obs_key + '*' + exp_key + '*' + comb_list[k] + '*IMI*')[0][18]
                combined_filename = 'g' + obs_key + 'OM' + mode_symbol + exp_key + 'CIMAGE0000.FIT'
                k = k + 1
        
        #parsing the list into a string...:
        exposure_image_list = str(exposure_image_list).strip('[]')
        exposure_image_list.replace(',', ' ')

        Message("... Combining sub-windows (Engineering-2 Mode data) ... ")
        highlighted_message("*", "... omcomb ...")

        cmd = "imagesets=\"" + exposure_image_list + "\"" + " outset=\"" + out_directory + '/' + combined_filename + "\""
        print('omcomb arguments: {}'.format(cmd))

        if subprocess.Popen('omcomb {}'.format(cmd), shell = True).wait():
            raise OSError('Failed while running omcomb.')

        Message('... Produced a combined image {}'.format(combined_filename))
        combined_image_was_produced = 1

    Message('... Running Tracking tasks ... ')
    i_win = 0


    while i_win < len(comb_list):
        if fnmatch.filter(list_of_files, orb_key + '*' + obs_key + '*' + exp_key + '*' + comb_list[i_win] + '*IMI*'):
            print(comb_list)
            highlighted_message('-', 'Window {}'.format(comb_list[i_win]))

            if combine and combined_image_was_produced:
                input_image_filename = out_directory + '/' + combined_filename
            else:
                input_image_filename = inp_dir + '/' + fnmatch.filter(list_of_files, orb_key + '*' + obs_key + '*' + exp_key + '*' + comb_list[i_win] + '*IMI*')[0]

            Message('Image file name: {}'.format(input_image_filename))

            filter_id = pyutils.get_key_word(input_image_filename, 'FILTER')

            if filter_id == 200 or filter_id == 1000:
                if filter_id == 200:
                    grism_id = 'Grism 2 (visual)'
                    Message('Filter value corresponds to {}'.format(grism_id))
                elif filter_id == 1000:
                    grism_id = 'Grism 1 (UV)'
                    Message('Filter value corresponds to {}'.format(grism_id))
                else:
                    Message('Filter value does not correspond to a grism.')


                raw_image = 'g' + input_image_filename
                s = mode_symbol

                out_flat_filename = "g" + obs_key + "OM" + s + exp_key + "FLAFLD" + comb_list[i_win][1] + "000.FIT"
                osw_list_intermediary_detect_filename = "p" + obs_key + "OM" + s + exp_key + "SWSRLI" + comb_list[i_win][1] + "000.FIT"
                second_osw_list_intermediary_detect_filename = "p" + obs_key + "OM" + s + exp_key + "SWSRLI" + comb_list[i_win][1] + "001.FIT"
                image_pps_product_filename = "p" + obs_key + "OM" + s + exp_key + "IMAGE_" + comb_list[i_win][1] + "000.FIT"
                intermediate_image_filename = "g" + obs_key + "OM" + s + exp_key + "IMAGEI" + comb_list[i_win][1] + "000.FIT"
                detectorCoordImageFileName = "g" + obs_key + "OM" + s + exp_key + "IMAGE_" + comb_list[i_win][1] + "000.FIT"
                skyCoordImageFileName = "g" + obs_key + "OM" + s + exp_key + "SIMAGE" + comb_list[i_win][1] + "000.FIT"
                rotatedImageFileName = "p" + obs_key + "OM" + s + exp_key + "RIMAGE" + comb_list[i_win][1] + "000.FIT"
                smoothRotatedImageFileName = "s" + obs_key + "OM" + s + exp_key + "IMAGE_" + comb_list[i_win][1] + "000.FIT"
                undistImageFileName = "u" + obs_key + "OM" + s + exp_key + "IMAGE_" + comb_list[i_win][1] + "000.FIT"

                unscatteredImageFileName = "g" + obs_key + "OM" + s + exp_key + "RIMNSC" + comb_list[i_win][1] + "000.FIT"
                bkgImageFileName = "g" + obs_key + "OM" + s + exp_key + "RIMBKG" + comb_list[i_win][1] + "000.FIT"

                modulo_8_product_filename = "g" + obs_key + "OM" + s + exp_key + "MOD8MP" + comb_list[i_win][1] + "000.FIT"
                sky_coord_pps_product_filename = "p" + obs_key + "OM" + s + exp_key + "SIMAGE" + comb_list[i_win][1] + "000.FIT"
                region_file = "p" + obs_key + "OM" + s + exp_key + "REGION" + comb_list[i_win][1] + "000.ASC"
                second_region_file = "p" + obs_key + "OM" + s + exp_key + "REGION" + comb_list[i_win][1] + "001.ASC"
                spectraRegionFileName = "p" + obs_key + "OM" + s + exp_key + "SPCREG" + comb_list[i_win][1] + "001.ASC"
                PPS_eventlist_file = "g" + obs_key + "OM" + s + exp_key + "EVLIST" + comb_list[i_win][1] + "000.FIT"

                spectraRegionPlotName = "p" + obs_key + "OM" + s + exp_key + "SPCREG" + comb_list[i_win][1] + "001.PS"
                spectraRegionPDF = "p" + obs_key + "OM" + s + exp_key + "SPCREG" + comb_list[i_win][1] + "001.PDF"
                # temporary files for testing purposes
                background_output = "tmp_background" + obs_key + "OM" + s + exp_key + "BACKGR" + comb_list[i_win][1] + ".FIT"
                signifimage_output = "tmp_signifimage" + obs_key + "OM" + s + exp_key + "SIGNIF" + comb_list[i_win][1] + ".FIT"
                grismOutFileName="p" + obs_key + "OM" + s + exp_key + "SPECTR" + comb_list[i_win][1] + "000.FIT"

                if extract_field_spectra:
                    Message("Available spectra of the field objects will be extracted (by request)")
                    out_spectra_list_filename = "p" + obs_key + "OM" + s + exp_key + "SPECLI" + comb_list[i_win][1] + "000.FIT"
                else:
                    Message('Extraction of the target object spectrum (no field spectra will be extracted)')
                    out_spectra_list_filename = "p" + obs_key + "OM" + s + exp_key + "SPECLI" + comb_list[i_win][1] + "000.FIT"

                Message('Running grism mode tasks...')

                highlighted_message('*', 'omgprep')

                arg_list = "set={}".format(input_image_filename) + " nphset={}/{}".format(inp_dir, non_periodic_hk_filename) +\
                    " pehset={}/{}".format(inp_dir, periodic_hk_filename) +\
                    " wdxset={}/{}".format(inp_dir, window_data_filename) +\
                    " modeset=4" + " outset={}/{}".format(out_dir, intermediate_image_filename)

                Message("omprep {}".format(arg_list))

                if os.system('omprep {}'.format(arg_list)):
                    highlighted_message('*', 'omprep has detected an error- this observation will not be processed')
                    return 'Error in omprep'

                highlighted_message('*', ' ... ommodmap ...')
                arg_list = "set={}/{}".format(out_dir, intermediate_image_filename) +\
                    " flatset={}".format(flat_field_filename) + " mod8product=yes" +\
                    " mod8set={}/{}".format(out_dir, modulo_8_product_filename) +\
                    " outset={}/{}".format(out_dir, detectorCoordImageFileName) +\
                    " outflatset={}/{}".format(out_dir, out_flat_filename) +\
                    " nsig={}".format(ommodmap_nsig) + " nbox={}".format(ommodmap_nbox) +\
                    " mod8correction={}".format(mod8correction)

                Message('ommodmap {}'.format(arg_list))
                if subprocess.Popen('ommodmap {}'.format(arg_list), shell = True).wait():
                    return 'Error while running ommodmap'


                highlighted_message('*', '... omgprep ...')
                omgprep_usecat = 'F'

                print("removeScatteredLight= {}\n".format(removeScatteredLight))
                if removeScatteredLight:
                    arg_list = "set={}/{}".format(out_dir, detectorCoordImageFileName) +\
                        " outset={}/{}".format(out_dir, rotatedImageFileName) +\
                        " undistset={}/{}".format(out_dir, undistImageFileName) +\
                        " removescatteredlight={}".format(removeScatteredLight) +\
                        " backgroundset={}/{}".format(out_dir, bkgImageFileName)
                else:
                    arg_list = "set={}/{}".format(out_dir, detectorCoordImageFileName) +\
                        " outset={}/{}".format(out_dir, rotatedImageFileName) +\
                        " undistset={}/{}".format(out_dir, undistImageFileName)

                Message('omgprep {}'.format(arg_list))
                if subprocess.Popen('omgprep {}'.format(arg_list), shell = True).wait():
                    highlighted_message('*', 'omgprep has found an error.')
                    return 'Error'

                level_image_file = 'LEVEL.FIT'

                highlighted_message('*', '... omdetect ...')
                arg_list = "nsigma={}".format(nsigma) +\
                    " set={}/{}".format(out_dir, rotatedImageFileName) +\
                    " regionfile={}/{}".format(out_dir, second_region_file) +\
                    " outset={}/{}".format(out_dir, second_osw_list_intermediary_detect_filename)

                Message('omdetect {}'.format(arg_list))
                if subprocess.Popen('omdetect {}'.format(arg_list), shell = True).wait():
                    return 'Error while running omdetect.'

                n_source = 0

                with open(out_dir + '/' + second_region_file) as region_file:
                    for line in region_file:
                        n_source = n_source + 1

                Message('Number of detected sources: {}'.format(n_source))

                highlighted_message('*', '... omatt ...')

                arg_list = "set=" +  out_dir + '/' + rotatedImageFileName +\
                    " sourcelistset=" +  out_dir + '/' + second_osw_list_intermediary_detect_filename +\
                    " ppsoswset=" + out_dir + '/' + skyCoordImageFileName + " usecat=F"

                Message('omatt {}'.format(arg_list))
                if subprocess.Popen('omatt {}'.format(arg_list), shell = True).wait():
                    return 'Error while running omatt'

                highlighted_message('*', '... omgrism ...')

                arg_list = "set={}/{}".format(out_dir, rotatedImageFileName) +\
                    " sourcelistset={}/{}".format(out_dir, second_osw_list_intermediary_detect_filename) +\
                    " outset={}/{}".format(out_dir, grismOutFileName) +\
                    " bkgoffsetleft={}".format(bkgOffsetLeft) +\
                    " bkgwidthleft={}".format(bkgWidthLeft) + " bkgoffsetright={}".format(bkgOffsetRight) +\
                    " bkgwidthright={}".format(bkgWidthRight) + " spectrumhalfwidth={}".format(spectrumHalfWidth) +\
                    " spectrumsmoothlength={}".format(spectrumsmoothlength) +\
                    " extractionmode={}".format(extractionmode) + " extractfieldspectra={}".format(extract_field_spectra) +\
                    " regionfile={}/{}".format(out_dir, second_region_file) + " spectraregionfile={}/{}".format(out_dir, spectraRegionFileName) +\
                    " outspectralistset={}".format(out_spectra_list_filename) + " addedregionfile={}".format(addedRegionFile)

                Message('omgrism {}'.format(arg_list))

                if subprocess.Popen('omgrism {}'.format(arg_list), shell = True).wait():
                    return 'Error while running omgrism.'

                spectrum_plot_file0 = "g" + obs_key + "OM" + s + exp_key + "SPECTR" + comb_list[i_win][1] + "000.PS"
                spectrum_PDF_file0 = "p" + obs_key + "OM" + s + exp_key + "SPECTR" + comb_list[i_win][1] + "000.PDF"

                highlighted_message('*', '... omgrismplot ...')
                Message('OMGRISM spectrum pps file: {}'.format(grismOutFileName))

                if os.path.exists(out_dir + '/' + grismOutFileName):
                    plotBinSize = 1
                    os.environ['PGPLOT_TYPE'] = 'ps'

                    arg_list = "set=" + out_dir + '/' + grismOutFileName +\
                        " scalebkgplot={}".format(scaleBkgPlot) + " binsize={}".format(plotBinSize) +\
                        " plotflux={}".format(plotFlux) +\
                        " plotfile=t.ps" + " spectraregionfile=" + out_dir + '/' + spectraRegionFileName +\
                        " regionplotfile=r.ps" + " rotatedimageset=" + out_dir + '/' + rotatedImageFileName

                    Message('omgrismplot {}'.format(arg_list))

                    if subprocess.Popen('omgrismplot {}'.format(arg_list), shell = True).wait():
                        return 'Error in omgrismplot'

                    os.rename('t.ps', out_dir + '/' + spectrum_plot_file0)
                    os.rename('r.ps', out_dir + '/' + spectraRegionPlotName)

                    print("spectraRegionFileName={}/{} \n".format(out_dir, spectraRegionFileName))
                    if os.path.isfile(out_dir + '/' + spectrum_plot_file0):
                        Message('Converting the spectrum PostScript file to PDF')
                        if subprocess.Popen('ps2pdf {}/{} {}/{}'.format(out_dir, spectrum_plot_file0, out_dir, spectraRegionPDF), shell = True).wait():
                            return 'Error while converting to pdf.'

                    else:
                        Message("The required spectrum PostScript file is missing")

                    if os.path.exists(out_dir + '/' + spectraRegionPlotName):
                        Message("Converting the spectrum region PostScript file to PDF")
                        if subprocess.Popen('ps2pdf {}/{} {}/{}'.format(out_dir, spectraRegionPlotName, out_dir, spectraRegionPDF), shell = True).wait():
                            return 'Error while converting to pdf.'
                    else:
                        Message("The spectrum region PostScript file is missing")
                else:
                    Message("File {} does not exist: no plot produced".format(grismOutFileName))
        else:
            Message('Image file for this window does not exist.')

        i_win = i_win + 1


def omgchain():
    """
    Main process for pyomgchain. Will start the whole chain.
    """

    py_version = 0.1

    if 'SAS_VERBOSITY' in os.environ:
        SAS_verbosity = os.environ['SAS_VERBOSITY']
    else:
        SAS_verbosity = '5'
        os.environ['SAS_VERBOSITY'] = '5'

    if int(SAS_verbosity) > 7:
        os.environ['SAS_VERBOSITY'] = '5'

    has_ff, ff_name, inp_dir, out_dir = set_up_directory_paths()

    inp_dir = inp_dir.rstrip()
    out_dir = out_dir.rstrip()

    periodic_file_flag = 0
    non_periodic_file_flag = 0
    flat_field_file_flag = 0

    task_name = 'OMGCHAIN'
    date = pyutils.date_time()
    Message(date)

    global ommodmap_nbox, ommodmap_nsig, spectrum_length_U
    global spectrum_length_V, src2_spectrum_U, src2_spectrum_V
    global image_smooth_box_idth, image_smooth_box_height
    ommodmap_nsig = 6
    ommodmap_nbox = 16

    spectrum_length_U = 1000.0
    spectrum_length_V = 400.0
    spectrumHalfWidth = -8.0

    src2_spectrum_U = 1100.0
    src2_spectrum_V = 800.0
    image_smooth_box_idth = 10.0
    image_smooth_box_height = 10.0

    create = 'Running SAS task {} V{} {}'.format(task_name, py_version, date)
    text1 = ' Input directory: {}'.format(inp_dir)
    text2 = ' Output directory: {}'.format(out_dir)
    highlighted_message('*', create + '  ' + comment + '  ' + text1 + ' ' + text2)

    if combine:
        Message("Engineering-2 Mode sub-windows (if they exist) will be combined")
    else:
        Message("Engineering-2 Mode windows will not be combined")

    global list_of_files

    list_of_files = glob.glob(inp_dir + '/*FIT')
    for i in range(0, len(list_of_files)):
        list_of_files[i] = list_of_files[i].replace(inp_dir + '/', '')

    global periodic_hk_filename
    global non_periodic_hk_filename

    periodic_hk_filename, non_periodic_hk_filename = check_for_house_keeping_files(list_of_files)

    global orb_list
    global obs_list
    global exp_list
    global win_list

    orb_list, obs_list, exp_list, win_list = fill_orb_obs_exp_dict(list_of_files)

    for orb_key in orb_list.keys():
        process_orbit(orb_key, has_ff, ff_name)
# end of omgchain


def fill_orb_obs_exp_dict(list_of_files):
    """
    Fill several dictionaries containing info of the orbit, the observation,
    the exposure and the window.
    Args:
        list_of_files: the list of FIT files to evaluate.
    
    Ouput: 
        orb_list, obs_list, exp_list, win_list: (orbit dictionary, obsid dictionary, exposure dicitonary, window dictionary).
    """

    orb_list = dict()
    obs_list = dict()
    exp_list = dict()
    win_list = dict()

    for i in list_of_files:

        if i.endswith('IMI.FIT'):
            print(i)
            orbit = i[0:4]
            obs = i[5:15]
            expo = i[19:22]
            win = i[22:24]

            orb_list.update({i[0:4] : i[0:4]})
            if obs in obs_list:
                if orbit not in obs_list[obs]:
                    obs_list[obs].append(orbit)
            else:
                obs_list.update({obs : [orbit]})
            if expo in exp_list:
                if obs not in exp_list[expo]:
                    exp_list[expo].append(obs)
            else:
                exp_list.update({expo : [obs]})
            if win in win_list:
                if expo not in win_list[win]:
                    win_list[win].append(expo)
            else:
                win_list.update({win : [expo]})

    return (orb_list, obs_list, exp_list, win_list)


def run(iparsdic):
    
    print(f'Executing {__file__} {iparsdic}')

    start = time.time()
    # gather params...
    global combine
    
    if iparsdic['combine'] == 'yes':
        combine = True
    else:
        combine = False

    global mod8correction, spectrumHalfWidth, spectrumsmoothlength, plot_bin_size, extract_field_spectra, extractionmode
    
    extractionmode = iparsdic['extractionmode']
    mod8correction = iparsdic['mod8correction']
    spectrumHalfWidth = iparsdic['spectrumhalfwidth']
    plot_bin_size = iparsdic['plotbinsize']
    extract_field_spectra = iparsdic['extractfieldspectra']

    global bkgOffsetLeft, bkgWidthLeft, bkgWidthRight, bkgOffsetRight
    spectrumsmoothlength = iparsdic['spectrumsmoothlength']
    bkgOffsetLeft = iparsdic['bkgoffsetleft']
    bkgWidthLeft = iparsdic['bkgwidthleft']
    bkgWidthRight = iparsdic['bkgwidthright']
    bkgOffsetRight = iparsdic['bkgoffsetright']
    
    global addedRegionFile, scaleBkgPlot, plotFlux, removeScatteredLight, nsigma
    plotFlux = iparsdic['plotflux']
    scaleBkgPlot = iparsdic['scalebkgplot']
    addedRegionFile = iparsdic['addedregionfile']
    removeScatteredLight = iparsdic['removescatteredlight']
    
    if removeScatteredLight == 'yes' or removeScatteredLight == 1:
        removeScatteredLight = True
    else:
        removeScatteredLight = False

    nsigma = iparsdic['nsigma']

    global comment, inpdirectory, outdirectory
    inpdirectory = iparsdic['inpdirectory']
    outdirectory = iparsdic['outdirectory']
    comment = iparsdic['comment']


    omgchain()

    stop = time.time()

    highlighted_message('*', 'pyomgchain ended. Total time: {} seconds.'.format(stop - start))
